from __future__ import annotations

import typing as t
from fastapi import HTTPException
from lzl.types import eproperty
from lzl.logging import logger
from lzl.api import aiohttpx
from pydantic.alias_generators import to_camel
from lzo.utils import timed_cache, Timer
from .models import (
    KeycloakUser,
    KeycloakGroup,
    KeycloakRealm,
    KeycloakClient,

    KeycloakWebhookCreate, 
    KeycloakWebhookObject
)

if t.TYPE_CHECKING:
    from .config import KeycloakSettings


_RealmToClients = {}


class KeycloakManager:
    def __init__(
        self, 
        url: t.Optional[str] = None, 
        client_id: t.Optional[str] = None, 
        username: t.Optional[str] = None, 
        password: t.Optional[str] = None,
        settings: t.Optional[KeycloakSettings] = None,
    ):
        """
        Initialize the KeycloakTokenManager.

        :param url: Keycloak server URL.
        :param client_id: Client ID for Keycloak.
        :param username: Username for Keycloak.
        :param password: Password for Keycloak.
        :param settings: Keycloak settings.
        """
        self._extra: t.Dict[str, t.Any] = {}
        if settings is None:
            from lzl.api.keycloak.registry import get_keycloak_settings
            settings = get_keycloak_settings()
        self.settings = settings
        self.url = url or self.settings.url
        self.master_realm = 'master'  # Always use master realm for token
        self.client_id = client_id or self.settings.client_id
        self.username = username or self.settings.username
        self.password = password or self.settings.password
        self._refresh_int = self.settings.refresh_realms_interval
        self._ts = Timer()
        self._index: t.Dict[str, t.List[KeycloakClient]] = {}

    @eproperty
    def token_url(self) -> str:
        """
        Returns the Keycloak token URL.
        """
        return f"{self.url}/realms/{self.master_realm}/protocol/openid-connect/token"

    @eproperty
    def api(self) -> aiohttpx.Client:
        """
        Returns the Keycloak API client.
        """
        from .auth import KeycloakAuth
        return aiohttpx.Client(
            base_url=self.url,
            auth=KeycloakAuth(self),
            preset_config='polling',
            follow_redirects=True,
            headers={
                'Content-Type': 'application/json',
            }
        )
    
    def _should_refresh_realms(self) -> bool:
        """
        Determines if the realms should be refreshed.
        This is a simple cache to avoid hitting Keycloak too often.
        """
        if self._ts.elapsed > self._refresh_int:
            logger.info(f"Refreshing Realms: {self._ts.elapsed} > {self._refresh_int}")
            self._ts = Timer()
            return True
        return False
        

    async def _apopulate_realms(self, realms: t.Optional[t.List[KeycloakRealm]] = None):
        """
        Populate the keycloak realms and realm clients
        """
        # logger.info("Populating Keycloak realms and clients")
        realms = realms or await self.get_realms()
        for realm in realms:
            self._index[realm.name] = await self.get_realm_clients(realm=realm.name, exclude_defaults=True)


    async def _adetermine_realm(self, realm: t.Optional[str] = None, client_id: t.Optional[str] = None) -> str:
        """
        Determines the realm
        """
        if realm: return realm
        if client_id:
            for r, clients in _RealmToClients.items():
                if client_id in clients:
                    return r

            if not self._index or self._should_refresh_realms(): await self._apopulate_realms()
            for r, clients in self._index.items():
                for c in clients:
                    if c.client_id == client_id or c.id == client_id or c.name == client_id:
                        return r
        raise ValueError(f"Unable to determine realm: {self._index} {client_id} {realm}")
    

    def _populate_realms(self, realms: t.Optional[t.List[KeycloakRealm]] = None):
        """
        Populate the keycloak realms
        """
        realms = realms or  self._get_realms()
        for realm in realms:
            self._index[realm.name] = self._get_realm_clients(realm=realm.name, exclude_defaults=True)
        
    
    def _determine_realm(self, realm: t.Optional[str] = None, client_id: t.Optional[str] = None) -> str:
        """
        Determines the realm
        """
        if realm: return realm
        if client_id:
            for r, clients in _RealmToClients.items():
                if client_id in clients:
                    return r
            if not self._index or self._should_refresh_realms(): self._populate_realms()
            for r, clients in self._index.items():
                for c in clients:
                    if c.client_id == client_id or c.id == client_id or c.name == client_id:
                        return r
        raise ValueError("Unable to determine realm")

    """
    User Methods
    """
    @t.overload
    async def list_users(
        self, 
        realm: t.Optional[str] = None, 
        client_id: t.Optional[str] = None, 
        email: t.Optional[str] = None,
        email_verified: t.Optional[bool] = None,
        enabled: t.Optional[bool] = None,
        exact: t.Optional[bool] = None,
        first: t.Optional[int] = None,
        username: t.Optional[str] = None,
        search: t.Optional[str] = None,
        q: t.Optional[str] = None,
        **kwargs
    ) -> t.List[KeycloakUser]:
        """
        List users in a Keycloak realm.

        :param realm: The realm to list users from.
        :param client_id: The client ID to list users for.
        :param email: The email address to filter users by.
        :param email_verified: Whether to filter users by email verification status.
        :param enabled: Whether to filter users by enabled status.
        :param exact: Whether to perform an exact match search.
        :param first: The index of the first user to return (for pagination).
        :param username: The username to filter users by.
        :param search: A search term to filter users by.
        :param q: A query string to filter users by.
        :param kwargs: Additional query parameters for filtering users.
        :return: A list of KeycloakUser objects.
        """
        ...

    async def list_users(self, realm: t.Optional[str] = None, client_id: t.Optional[str] = None, **kwargs) -> t.List[KeycloakUser]:
        """
        List users in a Keycloak realm.

        :param realm: The realm to list users from.
        :param client_id: The client ID to list users for.
        :param kwargs: Additional query parameters for filtering users.
        :return: A list of KeycloakUser objects.
        """
        realm = await self._adetermine_realm(realm, client_id)
        url_path = f"/admin/realms/{realm}/users"
        kwargs = {
            to_camel(k): v for k, v in kwargs.items() if v is not None
        }
        response = await self.api.aget(url_path, params=kwargs)
        if response.status_code != 200:
            raise HTTPException(status_code=500, detail=f"Failed to fetch users from Keycloak: {response.text}")
        return [KeycloakUser.model_validate(u, context={'realm': realm}) for u in response.json()]


    async def get_user(self, user_id: t.Optional[str] = None, username: t.Optional[str] = None, email: t.Optional[str] = None, realm: t.Optional[str] = None, client_id: t.Optional[str] = None) -> KeycloakUser:
        """
        Fetch a Keycloak user by their ID, username, or email.

        :param user_id: The ID of the user to fetch.
        :param username: The username of the user to fetch.
        :param email: The email of the user to fetch.
        :param realm: The realm to fetch the user from.
        :param client_id: The client ID to determine the realm.
        :return: A KeycloakUser object.
        """
        realm = await self._adetermine_realm(realm, client_id)
        if not user_id:
            users = await self.list_users(realm=realm, username=username, email=email)
            if not users: raise HTTPException(status_code=404, detail="User not found")
            if len(users) > 1: raise HTTPException(status_code=400, detail="Multiple users found")
            user = users[0]
            user_id = user.id
        response = await self.api.aget(f"/admin/realms/{realm}/users/{user_id}")
        if response.status_code != 200:
            raise HTTPException(status_code=500, detail=f"Failed to fetch user from Keycloak: {response.text}")
        return KeycloakUser.model_validate_json(response.text, context = {'realm': realm})


    async def update_user(self, user: 'KeycloakUser', realm: t.Optional[str] = None, client_id: t.Optional[str] = None) -> None:
        """
        Update a Keycloak user.

        :param user: The KeycloakUser object to update.
        :param realm: The realm to update the user in.
        :param client_id: The client ID to determine the realm.
        :return: None
        """
        realm = await self._adetermine_realm(realm or user.realm, client_id)
        url_path = f"/admin/realms/{realm}/users/{user.id}"
        user_data = user.model_dump(mode='json', by_alias=True)
        response = await self.api.aput(url_path, json=user_data)
        if response.status_code != 204:
            raise HTTPException(status_code=500, detail=f"Failed to update user in Keycloak: {response.text}")
        logger.info(f"User {user.id} updated successfully in Keycloak.")


    async def get_user_email(self, user_id: t.Optional[str] = None, username: t.Optional[str] = None, realm: t.Optional[str] = None, client_id: t.Optional[str] = None) -> str:
        """
        Fetch the email address of a user from Keycloak using their user ID or username.

        :param user_id: The ID of the user.
        :param username: The username of the user.
        :param realm: The realm to fetch the user from.
        :param client_id: The client ID to determine the realm.
        :return: The email address of the user.
        """
        user = await self.get_user(user_id=user_id, username=username, realm=realm, client_id=client_id)
        return user.email


    
    async def get_user_groups(self, user_id: str, realm: t.Optional[str] = None, client_id: t.Optional[str] = None) -> t.List[KeycloakGroup]:
        """
        Fetch the groups for a Keycloak user by their ID.
        
        :param user_id: The ID of the user.
        :param realm: The realm to fetch the user groups from.
        :param client_id: The client ID to determine the realm.
        :return: A list of KeycloakGroup objects.
        """
        realm = await self._adetermine_realm(realm, client_id)
        url_path = f"{self.url}/admin/realms/{realm}/users/{user_id}/groups"
        response = await self.api.aget(url_path)
        if response.status_code != 200:
            raise HTTPException(status_code=500, detail=f"Failed to fetch user groups from Keycloak: {response.text}")
        return [KeycloakGroup.model_validate(g, context={'realm': realm}) for g in response.json()]

    @t.overload
    async def get_user_group_names(
        self,
        user_id: str,
        realm: t.Optional[str] = None,
        client_id: t.Optional[str] = None
    ) -> t.List[str]:
        """
        Fetch the list of group names for a user from Keycloak Admin API.
        
        :param user_id: The ID of the user.
        :param realm: The realm to fetch the user groups from.
        :param client_id: The client ID to determine the realm.
        :return: A list of group names (strings).
        """
        ...

    async def get_user_group_names(self, user_id: str, realm: t.Optional[str] = None, client_id: t.Optional[str] = None) -> t.List[str]:
        """
        Fetch the list of group names for a user from Keycloak Admin API.
        Returns a list of group names (strings).
        """
        groups = await self.get_user_groups(user_id, realm, client_id)
        return [group.name for group in groups]
    
    """
    Realm Operations
    """

    def _get_realms(self, **kwargs) -> t.List[KeycloakRealm]:
        """
        Retrieves all realms from Keycloak.
        """
        response = self.api.get("/admin/realms")
        if response.status_code != 200:
            raise HTTPException(status_code=500, detail=f"Failed to fetch realms from Keycloak: {response.text}")
        return [KeycloakRealm.model_validate(r) for r in response.json()]

    async def get_realms(self, brief_representation: t.Optional[bool] = None, **kwargs) -> t.List[KeycloakRealm]:
        """
        Retrieves all realms from Keycloak.
                
        :param brief_representation: Whether to return a brief representation of the realms.
        :param kwargs: Additional query parameters.
        :return: A list of KeycloakRealm objects.
        """
        response = await self.api.aget("/admin/realms")
        if response.status_code != 200:
            raise HTTPException(status_code=500, detail=f"Failed to fetch realms from Keycloak: {response.text}")
        return [KeycloakRealm.model_validate(r) for r in response.json()]
    

    def _get_realm_clients(
        self,
        realm: str,
        exclude_defaults: t.Optional[bool] = True,
        **kwargs,
    ) -> t.List[KeycloakClient]:
        """
        Returns all the clients that belong to this realm.
        """
        response = self.api.get(f"/admin/realms/{realm}/clients")
        if response.status_code != 200:
            raise HTTPException(status_code=500, detail=f"Failed to fetch clients from Keycloak: {response.text}")
        clients = [KeycloakClient.model_validate(c) for c in response.json()]
        if exclude_defaults: clients = [c for c in clients if not c.is_default_client_id]
        return clients

    @t.overload
    async def get_realm_clients(
        self,
        realm: str,
        exclude_defaults: t.Optional[bool] = True,
        client_id: t.Optional[str] = None,
        first: t.Optional[int] = None,
        max: t.Optional[int] = None,
        search: t.Optional[bool] = None,
        viewable_only: t.Optional[bool] = None,
        **kwargs
    ) -> t.List[KeycloakClient]:
        """
        Returns all the clients that belong to this realm.
        
        :param realm: The realm to fetch clients from.
        :param exclude_defaults: Whether to exclude default clients.
        :param client_id: Filter by client ID.
        :param first: The first result to return (for pagination).
        :param max: The maximum number of results to return.
        :param search: Whether to search.
        :param viewable_only: Whether to return only viewable clients.
        :param kwargs: Additional query parameters.
        :return: A list of KeycloakClient objects.
        """
        ...

    async def get_realm_clients(
        self,
        realm: str,
        exclude_defaults: t.Optional[bool] = True,
        **kwargs,
    ) -> t.List[KeycloakClient]:
        """
        Returns all the clients that belong to this realm.
        """
        kwargs = {
            to_camel(k): v for k, v in kwargs.items() if v is not None
        }
        response = await self.api.aget(f"/admin/realms/{realm}/clients", params=kwargs)
        if response.status_code != 200:
            raise HTTPException(status_code=500, detail=f"Failed to fetch clients from Keycloak: {response.text}")
        # data = response.json()
        # logger.info(data)
        clients = [KeycloakClient.model_validate(c) for c in response.json()]
        if exclude_defaults:
            clients = [c for c in clients if not c.is_default_client_id]
        return clients
        
    """
    Webhook Operations
    """


    async def get_webhooks(self, realm: t.Optional[str] = None, client_id: t.Optional[str] = None) -> t.List[KeycloakWebhookObject]:
        """
        Fetch all webhooks from Keycloak.
        
        :param realm: The realm to fetch webhooks from.
        :param client_id: The client ID to determine the realm.
        :return: A list of KeycloakWebhookObject.
        """
        realm = await self._adetermine_realm(realm, client_id)
        url_path = f"/realms/{realm}/webhooks"
        response = await self.api.aget(url_path)
        if response.status_code != 200:
            raise HTTPException(status_code=500, detail=f"Failed to fetch webhooks from Keycloak: {response.text}")
        return [KeycloakWebhookObject.model_validate(w) for w in response.json()]

    async def get_webhook_id(
        self,
        url: t.Optional[str] = None,
        webhook_id: t.Optional[str] = None,
        realm: t.Optional[str] = None,
        client_id: t.Optional[str] = None
    ) -> str:
        """
        Fetch the ID of a webhook by its URL or ID.
        
        :param url: The URL of the webhook.
        :param webhook_id: The ID of the webhook.
        :param realm: The realm to search in.
        :param client_id: The client ID to determine the realm.
        :return: The webhook ID.
        """
        if webhook_id: return webhook_id
        if url:
            realm = await self._adetermine_realm(realm, client_id)
            webhooks = await self.get_webhooks(realm=realm, client_id=client_id)
            for webhook in webhooks:
                if webhook.url == url:
                    return webhook.id
            raise HTTPException(status_code=404, detail=f"Webhook with URL {url} not found.")
        raise ValueError("Either 'url' or 'webhook_id' must be provided to get a webhook ID.")

    async def get_webhook(
        self,
        url: t.Optional[str] = None,
        webhook_id: t.Optional[str] = None,
        realm: t.Optional[str] = None,
        client_id: t.Optional[str] = None,
        **kwargs,
    ) -> KeycloakWebhookObject:
        """
        Fetch an existing webhook.
        
        :param url: The URL of the webhook to fetch.
        :param webhook_id: The ID of the webhook to fetch.
        :param realm: The realm to fetch the webhook from.
        :param client_id: The client ID to determine the realm.
        :param kwargs: Additional parameters.
        :return: A KeycloakWebhookObject.
        """
        realm = await self._adetermine_realm(realm, client_id)
        webhook_id = await self.get_webhook_id(url=url, webhook_id=webhook_id, realm=realm, client_id=client_id)
        url_path = f"/realms/{realm}/webhooks/{webhook_id}"
        response = await self.api.aget(url_path)
        if response.status_code != 200:
            raise HTTPException(status_code=500, detail=f"Failed to fetch webhook from Keycloak: {response.text}")
        return KeycloakWebhookObject.model_validate(response.json())

    @t.overload
    async def create_webhook(
        self,
        url: str,
        enabled: t.Optional[bool] = True,
        events: t.Optional[t.List[str]] = None,
        secret: t.Optional[str] = None,
        realm: t.Optional[str] = None,
        client_id: t.Optional[str] = None,
        **kwargs,
    ) -> KeycloakWebhookObject:
        """
        Create a Keycloak webhook.

        :param url: The URL to send the webhook events to.
        :param enabled: Whether the webhook is enabled.
        :param events: List of events to subscribe to.
        :param secret: Optional secret for the webhook.
        :param realm: The realm to create the webhook in.
        :param client_id: The client ID to create the webhook for.
        """
        ...


    async def create_webhook(
        self,
        url: str,
        overwrite: t.Optional[bool] = False,
        realm: t.Optional[str] = None,
        client_id: t.Optional[str] = None,
        **kwargs,
    ) -> KeycloakWebhookObject:
        """
        Create a Keycloak webhook.

        :param url: The URL to send the webhook events to.
        :param enabled: Whether the webhook is enabled.
        :param events: List of events to subscribe to.
        :param secret: Optional secret for the webhook.
        :param realm: The realm to create the webhook in.
        :param client_id: The client ID to create the webhook for.
        """
        # We'll check if the webhook already exists
        realm = await self._adetermine_realm(realm, client_id)
        if not overwrite:
            webhooks = await self.get_webhooks(realm=realm, client_id=client_id)
            for webhook in webhooks:
                if webhook.url == url:
                    logger.info(f"Webhook already exists: {webhook.id} in realm |y|{realm}|e| for url: {url}")
                    return webhook
        kwargs['url'] = url
        data_object = KeycloakWebhookCreate.model_validate(kwargs)
        data = data_object.model_dump(mode='json', by_alias=True, exclude_none=True)
        url_path = f"/realms/{realm}/webhooks"
        response = await self.api.apost(url_path, json=data)
        if response.status_code != 201:
            raise HTTPException(status_code=500, detail=f"Failed to create webhook in Keycloak: {response.text} for {url} in {realm}")
        return await self.get_webhook(url=url, realm=realm, client_id=client_id)
        

    @t.overload
    async def update_webhook(
        self,
        url: t.Optional[str] = None,
        webhook_id: t.Optional[str] = None,
        enabled: t.Optional[bool] = None,
        events: t.Optional[t.List[str]] = None,
        secret: t.Optional[str] = None,
        realm: t.Optional[str] = None,
        client_id: t.Optional[str] = None,
        **kwargs
    ) -> KeycloakWebhookObject:
        """
        Update an existing webhook.
        
        :param url: The URL to send the webhook events to.
        :param webhook_id: The ID of the webhook to update.
        :param enabled: Whether the webhook is enabled.
        :param events: List of events to subscribe to.
        :param secret: Optional secret for the webhook.
        :param realm: The realm to update the webhook in.
        :param client_id: The client ID to determine the realm.
        :param kwargs: Additional parameters.
        :return: The updated KeycloakWebhookObject.
        """
        ...

    async def update_webhook(
        self,   
        url: t.Optional[str] = None,
        webhook_id: t.Optional[str] = None,
        realm: t.Optional[str] = None,
        client_id: t.Optional[str] = None,
        **kwargs,
    ) -> KeycloakWebhookObject:
        """
        Update an existing webhook

        :param url: The URL to send the webhook events to.
        :param webhook_id: The ID of the webhook to update.
        :param enabled: Whether the webhook is enabled.
        :param events: List of events to subscribe to.
        :param secret: Optional secret for the webhook.
        :param realm: The realm to update the webhook in.
        :param client_id: The client ID to determine the realm.
        :param kwargs: Additional parameters.
        :return: The updated KeycloakWebhookObject.
        """
        if not url and not webhook_id:
            raise ValueError("Either 'url' or 'webhook_id' must be provided to update a webhook.")
        realm = await self._adetermine_realm(realm, client_id)
        webhook_id = await self.get_webhook_id(url=url, webhook_id=webhook_id, realm=realm, client_id=client_id)
        url_path = f"/realms/{realm}/webhooks/{webhook_id}"
        kwargs['url'] = url
        data_object = KeycloakWebhookCreate.model_validate(kwargs)
        data = data_object.model_dump(mode='json', by_alias=True, exclude_none=True)
        response = await self.api.aput(url_path, json=data)
        if response.status_code != 204:
            raise HTTPException(status_code=500, detail=f"Failed to update webhook in Keycloak: {response.text}")
        return await self.get_webhook(url=url, realm=realm, client_id=client_id)
    

    async def delete_webhook(
        self,
        url: t.Optional[str] = None,
        webhook_id: t.Optional[str] = None,
        realm: t.Optional[str] = None,
        client_id: t.Optional[str] = None,
    ) -> t.Dict[str, str]:
        """
        Delete a Keycloak webhook.

        :param url: The URL of the webhook to delete.
        :param webhook_id: The ID of the webhook to delete.
        :param realm: The realm to delete the webhook from.
        :param client_id: The client ID to delete the webhook for.
        :return: A dictionary with a success message.
        """
        realm = await self._adetermine_realm(realm, client_id)
        webhook_id = await self.get_webhook_id(url=url, webhook_id=webhook_id, realm=realm, client_id=client_id)
        url_path = f"/realms/{realm}/webhooks/{webhook_id}"
        response = await self.api.adelete(url_path)
        if response.status_code != 204:
            raise HTTPException(status_code=500, detail=f"Failed to delete webhook in Keycloak: {response.text}")
        return {"detail": "Webhook deleted successfully"}


    async def _clear_all_webhooks(self, realm: t.Optional[str] = None, client_id: t.Optional[str] = None) -> None:
        """
        Delete all existing webhooks
        """
        realm = await self._adetermine_realm(realm, client_id)
        webhooks = await self.get_webhooks(realm=realm, client_id=client_id)
        for webhook in webhooks:
            await self.delete_webhook(webhook_id=webhook.id)



"""
asyncio_atexit: atexit for asyncio

borrowed from https://github.com/minrk/asyncio-atexit
"""

import asyncio
import inspect
import sys
import weakref
from functools import partial

__all__ = ["register", "unregister"]
__version__ = "1.0.1"

_registry = weakref.WeakKeyDictionary()

if sys.version_info < (3, 7):
    get_running_loop = asyncio.get_event_loop
else:
    get_running_loop = asyncio.get_running_loop


class _RegistryEntry:
    def __init__(self, loop):
        if not hasattr(loop, "_atexit_orig_close"):
            # avoid double-patching
            # weakrefs can get unresolved and then close called in __del__,
            # so this seems unavoidable
            loop._atexit_orig_close = loop.close
        try:
            self._close_ref = weakref.WeakMethod(loop._atexit_orig_close)
        except TypeError:
            # not everything can be weakref'd (Extensions such as uvloop).
            # Hold a regular reference _on the object_, in those cases
            self._close_ref = lambda: loop._atexit_orig_close
        self.callbacks = []

    def close(self):
        return self._close_ref()()


def register(callback, *, loop=None):
    """
    Register a callback for when the current event loop closes

    Like atexit.register, but run when the asyncio loop is closing,
    rather than process cleanup.

    `loop` may be specified as a keyword arg
    to attach to a non-running event loop.

    Allows coroutines to cleanup their resources.

    Callback will be passed no arguments.
    To pass arguments to your callback,
    use `functools.partial`.
    """
    entry = _get_entry(loop)
    entry.callbacks.append(callback)


def unregister(callback, *, loop=None):
    """
    Unregister a callback registered with asyncio_atexit.register

    `loop` may be specified as a keyword arg
    to attach to a non-running event loop.
    """

    entry = _get_entry(loop)
    # remove all instances of the callback
    while True:
        try:
            entry.callbacks.remove(callback)
        except ValueError:
            break


def _get_entry(loop=None):
    """Get the registry entry for an event loop"""
    if loop is None:
        loop = get_running_loop()
    _register_loop(loop)
    return _registry[loop]


def _register_loop(loop):
    """Patch an asyncio.EventLoop to support atexit callbacks"""
    if loop in _registry:
        return

    _registry[loop] = _RegistryEntry(loop)

    loop.close = partial(_asyncio_atexit_close, loop)


async def _run_asyncio_atexits(loop, callbacks):
    """Run asyncio atexit callbacks

    This runs in EventLoop.close() prior to actually closing the loop
    """
    for callback in callbacks:
        try:
            f = callback()
            if inspect.isawaitable(f):
                await f
        except Exception as e:
            print(
                f"Unhandled exception in asyncio atexit callback {callback}: {e}",
                file=sys.stderr,
            )


def _asyncio_atexit_close(loop):
    """Patched EventLoop.close method to run atexit callbacks

    prior to the unpatched close method.
    """
    entry = _get_entry(loop)
    if entry.callbacks:
        loop.run_until_complete(_run_asyncio_atexits(loop, entry.callbacks))
    entry.callbacks[:] = []
    return entry.close()


"""
Serialization Utilities
"""

import abc
import datetime
import dataclasses
import contextlib
from uuid import UUID
from enum import Enum
from lzl.logging import logger
from lzl.load import lazy_import
from lzl.types import BaseModel, Literal
from .defaults import default_pickle
from typing import Optional, Union, Any, Dict, List, Tuple, Callable, Type, Mapping, TypeVar, TYPE_CHECKING

try:
    import numpy as np
    np_version = np.__version__
    if np_version.startswith("2."):
        np_float_types = (np.float16, np.float32, np.float64)
    else:
        np_float_types = (np.float_, np.float16, np.float32, np.float64)
    np_int_types = (np.int_, np.intc, np.intp, np.int8, np.int16, np.int32, np.int64, np.uint8, np.uint16, np.uint32, np.uint64)
    
except ImportError:
    np = None


SerMode = Literal['auto', 'raw']

SerializableObject = TypeVar('SerializableObject')
serialization_class_registry: Dict[str, Type[SerializableObject]] = {}

_alias_schema_mapping: Dict[str, str] = {}

def is_primitive(value, exclude_bytes: Optional[bool] = False) -> bool:
    """
    Check if a value is a primitive type
    """
    if exclude_bytes and isinstance(value, bytes): return False
    return isinstance(value, (int, float, bool, str, bytes, type(None)))


def get_object_classname(obj: SerializableObject, is_type: Optional[bool] = False) -> str:
    """
    Get the classname of an object
    """
    if is_type: return f'{obj.__module__}.{obj.__name__}'
    return f"{obj.__class__.__module__}.{obj.__class__.__name__}"

def get_object_class(name: str) -> Type[SerializableObject]:
    """
    Get the class of an object
    """
    global serialization_class_registry
    if name not in serialization_class_registry:
        serialization_class_registry[name] = lazy_import(name)
    return serialization_class_registry[name]

def register_object_class(obj: SerializableObject, is_type: Optional[bool] = False) -> str:
    """
    Register the object class
    """
    global serialization_class_registry
    obj_class_name = get_object_classname(obj, is_type = is_type)
    if obj_class_name not in serialization_class_registry:
        serialization_class_registry[obj_class_name] = obj if is_type else obj.__class__
    return obj_class_name

def register_schema_mapping(schemas: Dict[str, str]):
    """
    Register the schema mapping
    """
    global _alias_schema_mapping
    _alias_schema_mapping.update(schemas)

def extract_model_dumps_kwargs(kwargs: Dict[str, Any]) -> Dict[str, Any]:
    """
    Extracts the model dumps kwargs
    """
    return {
        k: v for k,v in kwargs.items() if k in {
            'include', 
            'exclude', 
            'context', 
            'by_alias', 
            'exclude_unset',
            'exclude_defaults',
            'exclude_none',
            'warnings',
            'serialize_as_any',
        }
    }



def serialize_object(
    obj: SerializableObject,
    mode: Optional[SerMode] = 'auto',
    **kwargs
) -> Union[Dict[str, Any], List[Dict[str, Any]], Any]:
    # sourcery skip: extract-duplicate-method
    """
    Helper to serialize an object

    Args:
        obj: the object to serialize

    Returns:
        the serialized object in dict
        
        if not disable_nested_values:
        {
            "__type__": ...,
            "value": ...,
        }

        otherwise for JSON Objects:

        {
            "__type__": ...,
            ...,
        }

    """
    if obj is None: return None
    disable_nested_values: Optional[bool] = kwargs.get('disable_nested_values')

    if isinstance(obj, BaseModel) or hasattr(obj, 'model_dump'):
        obj_class_name = register_object_class(obj)
        obj_value = obj.model_dump(mode = 'json', round_trip = True, context = {'source': 'io', 'method': 'serializer'}, **extract_model_dumps_kwargs(kwargs))
        if mode == 'raw': return obj_value
        if disable_nested_values:
            return {
                "__type__": "pydantic",
                "__class__": obj_class_name,
                **obj_value,
            }
        return {
            "__type__": "pydantic",
            "__class__": obj_class_name,
            "value": obj_value,
        }

    # Move this to the top before primitives
    if np is not None:
        # if isinstance(obj, (np.int_, np.intc, np.intp, np.int8, np.int16, np.int32, np.int64, np.uint8, np.uint16, np.uint32, np.uint64)):
        if isinstance(obj, np_int_types):
            if mode == 'raw': return int(obj)
            obj_class_name = register_object_class(obj)
            return {
                "__type__": "numpy",
                "__class__": obj_class_name,
                "value": int(obj),
            }


        # if isinstance(obj, (np.float_, np.float16, np.float32, np.float64)):
        if isinstance(obj, np_float_types):
            if mode == 'raw': return float(obj)
            obj_class_name = register_object_class(obj)
            return {
                "__type__": "numpy",
                "__class__": obj_class_name,
                "value": float(obj),
            }


    if is_primitive(obj, exclude_bytes = True):
        return obj
    
    if isinstance(obj, type):
        obj_class_name = register_object_class(obj, is_type = True)
        if mode == 'raw': return obj_class_name
        return {
            "__type__": "type",
            "__class__": obj_class_name,
            "value": obj_class_name,
        }

    if isinstance(obj, (list, tuple)):
        return [serialize_object(item, mode = mode, **kwargs) for item in obj]

    if isinstance(obj, dict):
        if "__type__" in obj: 
            return obj['value'] if mode == 'raw' and obj.get('value') else obj
        return {key: serialize_object(value, mode = mode, **kwargs) for key, value in obj.items()}

    if isinstance(obj, (datetime.datetime, datetime.date, datetime.time)):
        if mode == 'raw': return obj.isoformat()
        return {
            "__type__": "datetime",
            "value": obj.isoformat(),
        }

    if isinstance(obj, datetime.timedelta):
        if mode == 'raw': return obj.total_seconds()
        return {
            "__type__": "timedelta",
            "value": obj.total_seconds(),
        }

    if isinstance(obj, dataclasses.InitVar) or dataclasses.is_dataclass(obj):
        if mode == 'raw': return dataclasses.asdict(obj)
        obj_class_name = register_object_class(obj)
        if disable_nested_values:
            return {
                "__type__": "dataclass",
                "__class__": obj_class_name,
                **dataclasses.asdict(obj),
            }
        return {
            "__type__": "dataclass",
            "__class__": obj_class_name,
            "value": dataclasses.asdict(obj),
        }

    if hasattr(obj, 'as_posix'):
        if mode == 'raw': return obj.as_posix()
        obj_class_name = register_object_class(obj)
        return {
            "__type__": "path",
            "__class__": obj_class_name,
            "value": obj.as_posix(),
        }

    if isinstance(obj, (bytes, bytearray)):
        if mode == 'raw': return obj
        return {
            "__type__": "bytes",
            "value": obj.hex(),
        }

    if isinstance(obj, (set, frozenset)):
        if mode == 'raw': return list(obj)
        return {
            "__type__": "set",
            "value": list(obj),
        }

    if isinstance(obj, Enum):
        if mode == 'raw': return obj.value
        obj_class_name = register_object_class(obj)
        return {
            "__type__": "enum",
            "__class__": obj_class_name,
            "value": obj.value,
        }

    if isinstance(obj, UUID):
        if mode == 'raw': return str(obj)
        return {
            "__type__": "uuid",
            "value": str(obj),
        }

    if hasattr(obj, 'serialize'):
        if mode == 'raw': return obj.serialize()
        obj_class_name = register_object_class(obj)
        return {
            "__type__": "serializable",
            "__class__": obj_class_name,
            "value": obj.serialize(),
        }

    if isinstance(obj, abc.ABC):
        logger.info(f'Pickle Serializing ABC Object: |r|({type(obj)}) {str(obj)[:1000]}', colored = True)
        # if mode == 'raw': raise TypeError(f"Cannot serialize object of type in raw mode {type(obj)}")
        obj_bytes = default_pickle.dumps(obj)
        if mode == 'raw': return obj_bytes
        return {
            "__type__": "pickle",
            "value": obj_bytes.hex(),
        }


    if hasattr(obj, "numpy"):  # Checks for TF tensors without needing the import
        if mode == 'raw': return obj.numpy().tolist()
        return {
            "__type__": "tensor",
            "value": obj.numpy().tolist(),
        }

    if hasattr(obj, 'tolist'): # Checks for torch tensors without importing
        if mode == 'raw': return obj.tolist()
        return {
            "__type__": "tensor",
            "value": obj.tolist(),
        }

    # Try one shot encoding objects
    # with contextlib.suppress(Exception):

    try:
        logger.info(f'Pickle Serializing Object: |r|({type(obj)}) {str(obj)[:1000]}', colored = True)
        obj_bytes = default_pickle.dumps(obj)
        if mode == 'raw': return obj_bytes
        return {
            "__type__": "pickle",
            "value": obj_bytes.hex(),
        }
    except Exception as e:

        logger.info(f'Error Serializing Object: |r|({type(obj)}) {e}|e| {str(obj)[:1000]}', colored = True)

    raise TypeError(f"Cannot serialize object of type {type(obj)}")


def deserialize_object(
    obj: Union[Dict[str, Any], List[Dict[str, Any]], Any], 
    schema_map: Optional[Dict[str, str]] = None, 
    allow_failed_import: Optional[bool] = False
) -> SerializableObject:
    # sourcery skip: extract-duplicate-method, low-code-quality
    """
    Deserialize an object

    Args:
        obj: the object to deserialize
    """
    if obj is None: return None
    if isinstance(obj, (list, tuple)):
        return [deserialize_object(item, schema_map = schema_map, allow_failed_import = allow_failed_import) for item in obj]

    if isinstance(obj, dict):
        if "__type__" not in obj:
            return {key: deserialize_object(value, schema_map = schema_map, allow_failed_import = allow_failed_import) for key, value in obj.items()}

        # obj_type = obj["__type__"]
        obj_type = obj.pop("__type__")
        if '__class__' in obj:
            if schema_map is not None and obj['__class__'] in schema_map:
                obj['__class__'] = schema_map[obj['__class__']]
            elif obj['__class__'] in _alias_schema_mapping:
                obj['__class__'] = _alias_schema_mapping[obj['__class__']]
        
        obj_class_type = obj.pop('__class__', None)
        if obj_type == "type":
            return get_object_class(obj_class_type)
        
        obj_value = obj["value"] if len(obj) == 1 and "value" in obj else obj
        if obj_type == "pydantic":
            try:
                obj_class = get_object_class(obj_class_type)
                # for k,v in obj_value.items():
                #     if not is_primitive(v):
                #         obj_value[k] = deserialize_object(v)
                if hasattr(obj_class, 'model_validate'):
                    return obj_class.model_validate(obj_value, context = {'source': 'io', 'method': 'deserializer'})
                return obj_class(**obj_value)
            except ImportError as e:
                if allow_failed_import:
                    return deserialize_object(obj_value, schema_map = schema_map, allow_failed_import = allow_failed_import)
                raise e

        if obj_type == "serializable":
            try:
                obj_class = get_object_class(obj_class_type)
                return obj_class(**obj_value)
            except ImportError as e:
                if allow_failed_import:
                    return deserialize_object(obj_value, schema_map = schema_map, allow_failed_import = allow_failed_import)
                raise e

        if obj_type == "numpy" and np is not None:
            # dtype = obj.get("__class__")
            dtype = obj_class_type
            if dtype: dtype = dtype.replace("numpy.", "")
            return np.array(obj_value, dtype = dtype)

        if obj_type == "pickle":
            try:
                obj_value = bytes.fromhex(obj_value)
                return default_pickle.loads(obj_value)
            except Exception as e:
                raise TypeError(f"Cannot deserialize object of type {obj_type}: {e}") from e

        if obj_type == "datetime":
            return datetime.datetime.fromisoformat(obj_value)

        if obj_type == "timedelta":
            return datetime.timedelta(seconds=obj_value)

        if obj_type == "dataclass":
            # obj_class_type = obj["__class__"]
            # if schema_map is not None and obj_class_type in schema_map:
            #     obj_class_type = schema_map[obj_class_type]

            obj_class = get_object_class(obj_class_type)
            return obj_class(**obj_value)

        if obj_type == "path":
            obj_class = get_object_class(obj_class_type)
            return obj_class(obj_value)

        if obj_type == "enum":
            obj_class = get_object_class(obj_class_type)
            return obj_class(obj_value)

        if obj_type == "uuid":
            return UUID(obj_value)

        if obj_type == "bytes":
            return bytes.fromhex(obj_value)

        if obj_type == "set":
            return set(obj_value)
        

        if obj_type == "tensor":
            return float(obj_value) if (isinstance(obj_value, (int, float)) or not np) else np.array(obj_value)

        raise TypeError(f"Cannot deserialize object of type {obj_type} - {obj_class_type} with value {obj_value}")

    if isinstance(obj, bytes):
        # Try to deserialize with pickle
        with contextlib.suppress(Exception):
            return default_pickle.loads(obj)

    return obj






from __future__ import annotations

import os
import abc
import signal
import pathlib
import asyncio
import contextlib
import multiprocessing
from lzl.proxied import ProxyObject
from lzl.logging import logger, Logger
from lzl.types import eproperty
from typing import Optional, List, TypeVar, Callable, Dict, Any, overload, Type, Union, TYPE_CHECKING


class ServerStateObject(abc.ABC):
    """
    The Server State
    """

    _extra: Dict[str, Any] = {}

    @eproperty
    def logger(self) -> 'Logger':
        """
        Returns the logger
        """
        return logger

    def start_debug_mode(self):
        """
        Enters Debug Mode
        """
        self.logger.warning('Entering Debug Mode. Sleeping Forever')
        import time
        import sys
        while True:
            try:
                time.sleep(900)
            except Exception as e:
                self.logger.error(e)
                break
        sys.exit(0)

    async def arun_until_complete(
        self,
        termination_file: Optional[str] = None,
    ):
        """
        Runs the event loop until complete
        """
        from .term import GracefulTermination
        watch = GracefulTermination()
        tmp_kill_file = pathlib.Path(termination_file) if termination_file is not None else None
        while not watch.kill_now:
            try: 
                await asyncio.sleep(1.0)
                if tmp_kill_file is not None and tmp_kill_file.exists():
                    self.logger.warning(f"Found termination file: {tmp_kill_file}")
                    break
            except KeyboardInterrupt:
                self.logger.warning("Keyboard Interrupt")
                break
            except Exception as e:
                self.logger.error(f"Error: {e}")
                break
        
        self.end_all_processes()
        await self.aclose_processes()



ServerState: ServerStateObject = ProxyObject(ServerStateObject)


"""Miscellaneous helper utilities used by :mod:`lzl.pool`."""

import asyncio
import contextlib


def is_in_async_loop() -> bool:
    """Return ``True`` when called from within a running event loop."""

    with contextlib.suppress(Exception):
        return asyncio.get_event_loop().is_running()
    return False



from __future__ import annotations

import datetime
from typing import Optional, Type, List, Dict, Any
from lzo.types import Field, eproperty
from lzl.api.openai.types.base import BaseResource, Permission
from lzl.api.openai.types.responses import BaseResponse
from lzl.api.openai.types.routes import BaseRoute
# from lazyops.types import eproperty, Field

# from async_openai.types.resources import BaseResource, Permission
# from async_openai.types.responses import BaseResponse
# from async_openai.types.routes import BaseRoute


__all__ = [
    'ModelData',
    'ModelObject',
    'ModelResponse',
    'ModelRoute',
]


class ModelData(BaseResource):
    id: str
    status: Optional[str] = None
    owned_by: Optional[str] = None
    created: Optional[datetime.datetime] = None
    updated_at: Optional[datetime.datetime] = None
    created_at: Optional[datetime.datetime] = None
    permission: Optional[List[Permission]] = Field(default_factory = list)
    root: Optional[str] = None
    parent: Optional[str] = None
    object: Optional[str] = 'model'
    capabilities: Optional[Dict[str, Any]] = Field(default_factory = dict)
    lifecycle_status: Optional[str] = None
    deprecation: Optional[Dict[str, Any]] = Field(default_factory = dict)


    @eproperty
    def model_age(self) -> Optional[datetime.datetime]:
        """
        Returns how long ago the model was created
        """
        if self.created:
            return datetime.datetime.now(tz = datetime.timezone.utc) - self.created


class ModelObject(BaseResource):
    model: Optional[str]


class ModelResponse(BaseResponse):
    data: Optional[List[ModelData]]
    data_model: Optional[Type[BaseResource]] = ModelData

    @eproperty
    def model_list(self) -> List[str]:
        """
        Returns a list of model IDs
        """
        return [model.id for model in self.data] if self.data and isinstance(self.data, list) else []


class ModelRoute(BaseRoute):
    input_model: Optional[Type[BaseResource]] = ModelObject
    response_model: Optional[Type[BaseResource]] = ModelResponse

    api_resource: Optional[str] = Field(default = 'models')

    # @eproperty
    # def api_resource(self):
    #     return 'models'
    
    @eproperty
    def create_enabled(self):
        """
        Returns whether the Create Route is Enabled
        """
        return True
    
    
    @eproperty
    def list_enabled(self):
        """
        Returns whether the List Route is Enabled
        """
        return False
    
    @eproperty
    def get_enabled(self):
        """
        Returns whether the Get Route is Enabled
        """
        return False

    def retrieve(
        self, 
        resource_id: str, 
        params: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> ModelResponse:
        """
        Retrieve a Single Model by Resource ID

        :param resource_id: The ID of the Resource to GET
        :param params: Optional Query Parameters
        """
        return super().retrieve(resource_id = resource_id, params = params, **kwargs)
    
    async def async_retrieve(
        self,
        resource_id: str,
        params: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> ModelResponse:
        """
        Retrieve a Single Model by Resource ID

        :param resource_id: The ID of the Resource to GET
        :param param
        """
        return await super().async_retrieve(resource_id = resource_id, params = params, **kwargs)


    def list(
        self, 
        params: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> ModelResponse:
        """
        List all available Models

        :param params: Optional Query Parameters
        """
        return super().list(params = params, **kwargs)
    
    async def async_list(
        self, 
        params: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> ModelResponse:
        """
        List all available Models

        :param params: Optional Query Parameters
        """
        return await super().async_list(params = params, **kwargs)

    






from __future__ import annotations

import typer
import typing as t
from lzl.ext.temporal.loop.main import cmd as worker_cmd

cmd = typer.Typer(no_args_is_help = True, help = "Temporal Commands")
cmd.add_typer(worker_cmd, name = "worker")




from __future__ import annotations

import os
import abc
import base64
import hashlib

from lzl import load
from lzl.api import aiohttpx
from lzl.types import eproperty
from lzl.logging import logger, Logger
from urllib.parse import urljoin, urlencode
from .token import OAuth2Token, AccessToken, IDToken
from . import errors
from typing import Optional, Union, Dict, Any, List, Type, Generator, TYPE_CHECKING

if load.TYPE_CHECKING:
    import jwt
else:
    jwt = load.LazyLoad("jwt", install_missing=True, install_options={'package': 'pyjwt'})

if TYPE_CHECKING:
    from lzl.io import PersistentDict
    from jwt import PyJWKClient
    from .config import BaseOAuth2Config
    from .user import OAuth2User, TokenT
    from .token import TokenTypes
    from ..manager import OAuth2Manager


class M2MTokenAuth(aiohttpx.Auth):
    """
    M2M Token Auth
    """
    def __init__(
        self,
        client: 'BaseOAuth2Client',
    ):
        """
        Initializes the M2M Token Auth
        """
        self.client = client
        self.token: Optional[OAuth2Token] = None
        self.token_key = f'm2m:{self.client.config.m2m_client_id}'


    def fetch_or_refresh_token(self):
        """
        Fetches the token
        """
        if self.token is None and self.client.data.contains(self.token_key):
            self.token = self.client.data.get(self.token_key)
        
        if self.token is None or self.token.is_expired:
            token_data = self.client.fetch_m2m_token()
            self.token = self.client.parse_token_payload(token_data)
            self.client.data.set(self.token_key, self.token)
            logger.info(f'Refreshed M2M Token for |g|{self.token_key}|e|', colored = True, prefix = self.client.name)
        

    @property
    def authorization(self) -> str:
        """
        Returns the authorization
        """
        self.fetch_or_refresh_token()
        return self.token.access_token.value


    def auth_flow(self, request: aiohttpx.Request) -> Generator[aiohttpx.Request, aiohttpx.Response, None]:
        """
        Execute the authentication flow.

        To dispatch a request, `yield` it:

        ```
        yield request
        ```

        The client will `.send()` the response back into the flow generator. You can
        access it like so:

        ```
        response = yield request
        ```

        A `return` (or reaching the end of the generator) will result in the
        client returning the last response obtained from the server.

        You can dispatch as many requests as is necessary.
        """
        request.headers['Authorization'] = f'Bearer {self.authorization}'
        yield request


class BaseOAuth2Client(abc.ABC):
    """
    Base Auth Client
    """
    name: Optional[str] = None
    _extra: Dict[str, Any] = {}

    
    def __init__(
        self, 
        manager: 'OAuth2Manager',
        config: Optional['BaseOAuth2Config'] = None,
        **kwargs,
    ):
        """
        Initializes the OAuth Client
        """
        self.mtg = manager
        self.config = config if config is not None else self.config_class(**kwargs)
        if self.config.name: self.name = self.config.name
        elif self.name: self.config.name = self.name
        self.init_oauth_config(**kwargs)
        self.post_init(**kwargs)


    """
    Initialization Methods
    """

    def init_oauth_config(self, **kwargs):
        """
        Initialize the config
        """
        pass


    def post_init(self, **kwargs):
        """
        Post Init
        """
        pass


    """
    Properties
    """

    @property
    def data(self) -> 'PersistentDict[str, OAuth2Token]': 
        """
        Returns the data
        """
        if 'data' not in self._extra:
            from lzl.io.persistence import PersistentDict
            base_key = self.config.cache_base_key or \
                f'lzl.oauth2.client.{self.name}.{self.config.client_id}'
            
            if not self.config.cache_config:
                self.config.cache_config = {
                    'serializer': 'json',
                    'url': os.getenv('OAUTH_REDIS_URL'),
                }
            elif 'serializer' not in self.config.cache_config:
                self.config.cache_config['serializer'] = 'json'
            
            self._extra['data'] = PersistentDict(
                base_key = base_key,
                expiration = self.config.cache_expiration, # 3 months
                **self.config.cache_config,
            )
        return self._extra['data']

    @property
    def refresh_data(self) -> 'PersistentDict[str, str]':
        """
        Returns the refresh data
        """
        if 'refresh_data' not in self._extra:
            self._extra['refresh_data'] = self.data.get_child('refresh')
        return self._extra['refresh_data']
    
    @property
    def user_data(self) -> 'PersistentDict[str, Any]':
        """
        Returns the user info data
        """
        if 'user_data' not in self._extra:
            self._extra['user_data'] = self.data.get_child('users')
        return self._extra['user_data']

    @eproperty
    def client(self) -> 'aiohttpx.Client':
        """
        Returns the aiohttpx client
        """
        return aiohttpx.Client()

    @property
    def m2m_token_auth(self) -> Optional[M2MTokenAuth]:
        """
        Returns the m2m token auth
        """
        if self.config.m2m_enabled and 'm2m_token_auth' not in self._extra:
            self._extra['m2m_token_auth'] = M2MTokenAuth(
                client = self,
            )
        return self._extra.get('m2m_token_auth')


    @property
    def m2m_client(self) -> Optional[aiohttpx.Client]:
        """
        Returns the M2M client
        """
        if self.config.m2m_enabled and 'm2m_client' not in self._extra:
            self._extra['m2m_client'] = aiohttpx.Client(
                base_url = self.config.issuer_url,
                auth = self.m2m_token_auth,
            )
        return self._extra.get('m2m_client')

    @eproperty
    def config_class(self) -> Type['BaseOAuth2Config']:
        """
        Returns the Config Class
        """
        raise NotImplementedError

    @eproperty
    def jwt_client(self) -> Optional['PyJWKClient']:
        """
        Returns the jwt client
        """
        return jwt.PyJWKClient(self.config.jwt_url)

    @eproperty
    def logger(self) -> 'Logger':
        """
        Returns the logger
        """
        return logger
    
    """
    Primary Auth Methods
    """

    def get_auth_url(
        self, 
        redirect_url: str,
        scope: Optional[str] = None,
        redirect_params: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> str:
        """
        Returns the auth url
        """
        if redirect_params:
            redirect_url = self.encode_params_to_url(redirect_params, redirect_url)
        params = self.config.get_auth_params(redirect_url = redirect_url, scope = scope, **kwargs)
        return self.encode_params_to_url(params, self.config.authorization_endpoint)

    def fetch_token(
        self,
        code: str,
        redirect_url: str,
        **kwargs,
    ) -> Optional[Dict[str, Any]]:
        """
        Fetches the token
        """
        params = self.config.get_fetch_token_params(code = code, redirect_url = redirect_url, **kwargs)
        response = self.client.post(
            url = self.config.token_endpoint,
            headers = self.config.get_headers('token'),
            data = params,
        )
        return response.json()
    
    async def afetch_token(
        self,
        code: str,
        redirect_url: str,
        **kwargs,
    ) -> Optional[Dict[str, Any]]:
        """
        Fetches the token
        """
        params = self.config.get_fetch_token_params(code = code, redirect_url = redirect_url, **kwargs)
        response = await self.client.async_post(
            url = self.config.token_endpoint,
            headers = self.config.get_headers('token'),
            data = params,
        )
        response.raise_for_status()
        return response.json()
    
    def refresh_token(
        self,
        refresh_token: str,
        **kwargs,
    ) -> Optional[Dict[str, Any]]:
        """
        Refreshes the token
        """
        params = self.config.get_refresh_token_params(refresh_token = refresh_token, **kwargs)
        response = self.client.post(
            url = self.config.token_endpoint,
            headers = self.config.get_headers('token'),
            data = params,
        )
        response.raise_for_status()
        return response.json()
    
    async def arefresh_token(
        self,
        refresh_token: str,
        **kwargs,
    ) -> Optional[Dict[str, Any]]:
        """
        Refreshes the token
        """
        params = self.config.get_refresh_token_params(refresh_token = refresh_token, **kwargs)
        response = await self.client.async_post(
            url = self.config.token_endpoint,
            headers = self.config.get_headers('token'),
            data = params,
        )
        response.raise_for_status()
        return response.json()


    async def arefresh_oauth_token(
        self,
        oauth_token: OAuth2Token,
        refresh_token: Optional[str] = None,
        **kwargs,
    ) -> Optional[OAuth2Token]:
        """
        Refreshes the oauth token
        """
        if not refresh_token:
            refresh_token = oauth_token.refresh_token or \
                await self.refresh_data.aget(oauth_token.identity)
        try:
            refreshed_token = await self.arefresh_token(refresh_token = refresh_token, **kwargs)
            self.logger.info(f'Refreshed OAuth Token for |g|{oauth_token.identity}|e|', colored = True, prefix = self.name)
        
        except Exception as e:
            self.logger.warning(f'Error refreshing token: {e} - Deleting {oauth_token.identity}', prefix = self.name)
            await self.data.adelete(oauth_token.identity)
            await self.refresh_data.adelete(oauth_token.identity)
            return None
        
        refreshed_token = self.parse_token_payload(refreshed_token)
        if not refreshed_token.refresh_token: refreshed_token.refresh_token = refresh_token
        await self.data.aset(refreshed_token.identity, refreshed_token)
        return refreshed_token
    
    
    def get_login_url(
        self, 
        additional_params: Optional[Dict[str, Any]] = None, 
        state: Optional[str] = None
    ) -> str:
        """
        Returns the login url
        """
        pass


    """
    FastAPI Handler Methods
    """
    def get_token_from_payload(self, payload: Dict[str, Any], token: str) -> 'TokenT':
        """
        Returns the token type from the payload
        """
        # This is likely a oauth2 token
        if payload.get('access_token'): 
            return self.config.oauth2_token_class.from_payload(payload = payload, token = token, provider = self.name)
        # Id Token usually has `picture` or `at_hash` in the payload
        if payload.get('picture') or payload.get('at_hash'): 
            return self.config.id_token_class.from_payload(payload = payload, token = token, provider = self.name)
        return self.config.access_token_class.from_payload(payload = payload, token = token, provider = self.name)

    def parse_authorization_token(
        self,
        token: str,
        options: Dict[str, bool] = None,
        raise_errors: Optional[bool] = None,
        **kwargs,
    ) -> Optional[Union[AccessToken, IDToken]]:
        """
        Parses the authorization token
        """
        try:
            payload = self.decode_jwt_token(token, options = options, **kwargs)
            return self.get_token_from_payload(payload, token = token)
        except jwt.ExpiredSignatureError as e:
            logger.warning(f'Expired Signature: {e} - {token}')
            raise errors.ExpiredTokenError(f'Expired Authorization Token - {token}', e) from e
        except Exception as e:
            logger.warning(f'Error Parsing Token: {e}')
            if raise_errors: raise errors.InvalidTokenError(f'Invalid Authorization Token - {token}', e) from e
            return None
    
    async def aparse_authorization_token(
        self,
        token: str,
        options: Dict[str, bool] = None,
        raise_errors: Optional[bool] = None,
        **kwargs,
    ) -> Optional[Union[AccessToken, IDToken]]:
        """
        Parses the authorization token
        """
        try:
            payload = await self.adecode_jwt_token(token, options = options, **kwargs)
            return self.get_token_from_payload(payload, token = token)
        except jwt.ExpiredSignatureError as e:
            logger.warning(f'Expired Signature: {e} - {token}')
            raise errors.ExpiredTokenError(f'Expired Authorization Token - {token}', e) from e
        
        except errors.OAuth2Exception as e:
            raise e
        
        except Exception as e:
            logger.warning(f'Error Parsing Token: {e}')
            if raise_errors: raise errors.InvalidTokenError(f'Invalid Authorization Token - {token}', e) from e
            return None
    
    """
    Session / API Key Methods
    """

    async def afetch_token_for_identity(
        self,
        identity: str,
        raise_errors: Optional[bool] = None,
        **kwargs,
    ) -> Optional[OAuth2Token]:
        """
        Fetches the token for the identity
        """
        if not await self.data.acontains(identity):
            logger.warning(f'No Token for Identity: {identity}')
            if raise_errors: raise errors.InvalidIdentityError(f'No Token for Identity: {identity}')
            return None
        
        oauth_token = await self.data.aget(identity)
        if oauth_token.is_expired:
            logger.warning(f'Token for Identity is Expired: {identity}')
            oauth_token = await self.arefresh_oauth_token(oauth_token)
            if oauth_token is None:
                logger.warning(f'Unable to refresh token for Identity: {identity}')
                if raise_errors: raise errors.InvalidIdentityError(f'Unable to refresh token for Identity: {identity}')
                return None
        return oauth_token
    
    """
    User Info Methods
    """

    def validate_user(
        self,
        user: 'OAuth2User',
        **kwargs,
    ) -> 'OAuth2User':
        """
        Validates the user
        """
        pass

    def retrieve_user_info(
        self,
        # user_id: str,
        user: 'OAuth2User',
        **kwargs,
    ) -> Optional[Dict[str, Any]]:
        """
        Retrieves the user info
        """
        pass

    def fetch_user_info(
        self,
        # user_id: str,
        user: 'OAuth2User',
        refresh: Optional[bool] = None,
        **kwargs,
    ) -> Optional[Dict[str, Any]]:
        """
        Fetches the user info
        """
        self.validate_user(user, **kwargs)
        if not user.provider_id: return
        if refresh or not self.user_data.contains(user.provider_id):
            user_info = self.retrieve_user_info(user.provider_id, **kwargs)
            if user_info: self.user_data.set(user.provider_id, user_info)
        else:
            user_info = self.user_data.get(user.provider_id)
        return user_info
    
    async def avalidate_user(
        self,
        user: 'OAuth2User',
        **kwargs,
    ) -> 'OAuth2User':
        """
        Validates the user
        """
        pass

    async def aretrieve_user_info(
        self,
        # user_id: str,
        user: 'OAuth2User',
        **kwargs,
    ) -> Optional[Dict[str, Any]]:
        """
        Retrieves the user info
        """
        pass

    async def afetch_user_info(
        self,
        # user_id: str,
        user: 'OAuth2User',
        refresh: Optional[bool] = None,
        **kwargs,
    ) -> Optional[Dict[str, Any]]:
        """
        Fetches the user info
        """
        await self.avalidate_user(user, **kwargs)
        if not user.provider_id: return
        if refresh or not await self.user_data.acontains(user.provider_id):
            user_info = await self.aretrieve_user_info(user, **kwargs)
            if user_info: await self.user_data.aset(user.provider_id, user_info)
        else:
            user_info = await self.user_data.aget(user.provider_id)
        return user_info



    """
    Helper Methods
    """

    def decode_jwt_token(
        self,
        token: str,
        options: Optional[Dict[str, bool]] = None,
        disable_signing_key: Optional[bool] = None,
        **kwargs,
    ) -> Optional[Dict[str, Any]]:
        """
        Decodes the jwt token
        """
        if options is None: 
            options = {
                'verify_exp': True,
                'verify_iat': True,
                'verify_aud': 'audience' in kwargs,
                'verify_signature': not disable_signing_key,
            }
        signing_key = "" if disable_signing_key else self.jwt_client.get_signing_key_from_jwt(token).key
        return jwt.decode(
            token,
            signing_key, 
            algorithms=['RS256'],
            issuer = self.config.issuer_url,
            **kwargs,
            options = options,
        )
    
    async def adecode_jwt_token(
        self,
        token: str,
        options: Optional[Dict[str, bool]] = None,
        disable_signing_key: Optional[bool] = None,
        **kwargs,
    ) -> Optional[Dict[str, Any]]:
        """
        Decodes the jwt token
        """
        return self.decode_jwt_token(token, options = options, disable_signing_key = disable_signing_key, **kwargs)

    def parse_jwt_token(
        self,
        jwt_token: str,
        kind: Optional[str] = None,
    ) -> Optional[Dict[str, Any]]:
        """
        Parses the jwt token
        """
        signing_key = self.jwt_client.get_signing_key_from_jwt(jwt_token)
        return jwt.decode(
            jwt_token,
            signing_key.key, 
            algorithms=['RS256'],
            issuer = self.config.issuer_url,
            options = {
                'verify_signature': True,
                'verify_exp': True,
                'verify_aud': False,
            }
        )
    

    def parse_authorization_token(
        self,
        authorization: str,
        audiences: Optional[List[str]] = None,
        verify_exp: bool = True,
    ) -> Optional[Dict[str, Any]]:
        """
        Parses the authorization token
        """
        return jwt.decode(
            authorization, 
            self.jwt_client.get_signing_key_from_jwt(authorization).key, 
            algorithms = ['RS256'],
            audience = audiences,
            issuer = self.config.jwt_issuer_url,
            options={
                'verify_signature': True,
                'verify_exp': verify_exp,
                'verify_aud': bool(audiences),
            }
        )


    def parse_token_payload(
        self,
        token: Dict[str, Any],
    ) -> Optional[OAuth2Token]:
        """
        Parses the token payload
        """
        auth_token = self.config.oauth2_token_class(**token)
        if auth_token.access_token:
            if decoded := self.parse_jwt_token(auth_token.access_token, kind = 'access_token'):
                auth_token.set_access_token(decoded)
        if auth_token.id_token:
            if decoded := self.parse_jwt_token(auth_token.id_token, kind = 'id_token'):
                auth_token.set_id_token(decoded)
        if auth_token.refresh_token and auth_token.identity:
            self.refresh_data[auth_token.identity] = auth_token.refresh_token
        auth_token.value = token
        auth_token.set_provider(self.name)
        return auth_token
    
    async def aparse_token_from_callback(
        self,
        token: str,
        **kwargs,
    ) -> Optional[OAuth2Token]:
        """
        Parses the token from the callback
        """
        oauth_token = self.parse_token_payload(token)
        await self.data.aset(oauth_token.identity, oauth_token)
        return oauth_token


    async def amerge_stored_token(
        self,
        token: 'OAuth2Token',
        **kwargs,
    ) -> 'OAuth2Token':
        """
        Merges the stored token
        """
        if await self.data.acontains(token.identity):
            stored_token = await self.data.aget(token.identity)
            token.merge(stored_token)
            self.logger.info(token, prefix = self.name)
        return token


    async def set_user_data(self, user: 'OAuth2User', **kwargs) -> None:
        """
        Sets the user data
        """
        pass


    """
    Static Methods
    """

    @staticmethod
    def create_code_challenge(secret_key: str) -> str:
        """
        Creates a code challenge
        """
        return base64.urlsafe_b64encode(hashlib.sha256(secret_key.encode()).digest()).decode().rstrip('=')


    @staticmethod
    def encode_params_to_url(params: Dict[str, Any], url: str) -> str:
        """
        Encodes the params to a url
        """
        return f'{url}?{urlencode(params)}'
    

    """
    M2M Methods
    """


    def fetch_m2m_token(
        self,
        audience: Optional[str] = None,
    ) -> Optional[Dict[str, Any]]:
        """
        Fetches the token
        """
        params = self.config.get_m2m_token_params(audience = audience)
        response = self.client.post(
            url = self.config.token_endpoint,
            headers = self.config.get_headers('token'),
            data = params,
        )
        response.raise_for_status()
        return response.json()
    

    async def afetch_m2m_token(
        self,
        audience: Optional[str] = None,
    ) -> Optional[Dict[str, Any]]:
        """
        Fetches the token
        """
        params = self.config.get_m2m_token_params(audience = audience)
        response = await self.client.async_post(
            url = self.config.token_endpoint,
            headers = self.config.get_headers('token'),
            data = params,
        )
        response.raise_for_status()
        return response.json()
    

    """
    M2M App Methods
    """


    async def app_retrieve_existing_domains(
        self,
        **kwargs,
    ) -> Optional[List[str]]:
        """
        Retrieves the existing app domains
        """
        pass

    async def app_add_authorized_domains(
        self,
        domains: List[str],
        **kwargs,
    ) -> None:
        """
        Adds the authorized app domains
        """
        pass

    async def app_retrieve_existing_callbacks(
        self,
        **kwargs,
    ) -> Optional[List[str]]:
        """
        Retrieves the existing app callbacks
        """
        pass

    async def app_add_authorized_callbacks(
        self,
        callback_urls: List[str],
        **kwargs,
    ) -> None:
        """
        Adds the authorized callbacks
        """
        pass

    async def app_retrieve_existing_logouts(
        self,
        **kwargs,
    ) -> Optional[List[str]]:
        """
        Retrieves the existing app logout urls
        """
        pass

    async def app_add_authorized_logouts(
        self,
        logout_urls: List[str],
        **kwargs,
    ) -> None:
        """
        Adds the authorized logouts
        """
        pass

    async def app_authorize_callbacks(
        self,
        domains: Union[str, List[str]] = None,
        path: Optional[str] = None,
    ) -> None:
        """
        Authorize Callbacks
        """
        if not self.m2m_client: return
        if self.config.authorize_disabled: return
        if not domains: domains = self.mtg.config.domains
        elif isinstance(domains, str): domains = [domains]
        if not path: 
            path = f'{self.mtg.routes["callback"]}/{self.name}'
            # path = f'/callback/{self.name}'
        if not domains: return
        callback_urls = [
            urljoin(domain, path)
            for domain in domains
        ]
        existing_callbacks = await self.app_retrieve_existing_callbacks()
        if existing_callbacks:
            callback_urls = [url for url in callback_urls if url not in existing_callbacks]

        if not callback_urls: return
        await self.app_add_authorized_callbacks(callback_urls)


    async def app_authorize_logouts(
        self,
        domains: Union[str, List[str]] = None,
        path: Optional[str] = None,
    ) -> None:
        """
        Authorizes the logout domains
        """
        if not self.m2m_client: return
        if self.config.authorize_disabled: return
        if not domains: domains = self.mtg.config.domains
        elif isinstance(domains, str): domains = [domains]
        if not path: 
            path = f'{self.mtg.routes["logout"]}/{self.name}'
            # path = f'/logout/{self.name}'
        if not domains: return
        logout_urls = [
            urljoin(domain, path)
            for domain in domains
        ]
        existing_logouts = await self.app_retrieve_existing_logouts()
        if existing_logouts:
            logout_urls = [url for url in logout_urls if url not in existing_logouts]
        if not logout_urls: return
        await self.app_add_authorized_logouts(logout_urls)

    async def app_authorize_domains(
        self,
        domains: Union[str, List[str]] = None,
        # domains: Optional[List[str]] = None,
    ) -> None:
        """
        Authorizes the app domains
        """
        if not self.m2m_client: return
        if self.config.authorize_disabled: return
        if not domains: domains = self.mtg.config.domains
        elif isinstance(domains, str): domains = [domains]
        existing_domains = await self.app_retrieve_existing_domains()
        if existing_domains:
            domains = [domain for domain in domains if domain not in existing_domains]

        if not domains: return
        await self.app_add_authorized_domains(domains)

    async def app_authorize_with_provider(
        self,
        domains: Union[str, List[str]] = None,
        callback_path: Optional[str] = None,
        logout_path: Optional[str] = None,
    ) -> None:
        """
        Authorizes the app with the provider
        """
        if not self.m2m_client: return
        if self.config.authorize_disabled: return
        if not domains: domains = self.mtg.config.domains
        elif isinstance(domains, str): domains = [domains]
        if not callback_path:  callback_path = f'{self.mtg.routes["callback"]}/{self.name}'
        if not logout_path: logout_path = f'{self.mtg.routes["logout"]}/{self.name}'
        await self.app_authorize_callbacks(domains = domains, path = callback_path)
        await self.app_authorize_logouts(domains = domains, path = logout_path)
        await self.app_authorize_domains(domains = domains)


    """
    Updating Users
    """



from __future__ import annotations

"""
Remote Object Storage Backend Dict-Like Persistence
"""

import concurrent.futures
from lzl.pool import ThreadPool
from lzo.types import Literal, eproperty
from lzl.io.file import File
from typing import TypeVar, Generic, Any, Dict, Optional, Union, Tuple, Iterable, List, Type, Callable, Generator, AsyncGenerator, TYPE_CHECKING
from ..base import BaseStatefulBackend, SchemaType, create_unique_id, logger
from .expirations import FileExpirationBackend, RedisExpirationBackend

if TYPE_CHECKING:
    from lzl.io.file import FileLike

# TODO: add caching support for lookups
_logged_backend: bool = False

def _display_backend(backend: str):
    """
    Displays the backend
    """
    global _logged_backend
    if _logged_backend: return
    logger.info(f'Using Object Storage Expiration Backend: `|g|{backend}|e|`', colored = True)
    _logged_backend = True


class ObjStorageStatefulBackend(BaseStatefulBackend):
    """
    Implements an Object Storage Stateful Backend

    Supports:
    - Local File System
    - S3
    - R2

    Implements either filebased expiration or redis expiration
    """
    name: Optional[str] = "objstore"
    expiration: Optional[int] = None
    file_ext: Optional[str] = None
    file_pre: Optional[str] = None
    auto_delete_invalid: Optional[bool] = False

    def __init__(
        self,
        base_key: Union[str, 'File'],
        name: Optional[str] = None,
        expiration: Optional[int] = None,
        async_enabled: Optional[bool] = False,
        file_ext: Optional[str] = None,
        file_pre: Optional[str] = None,
        auto_delete_invalid: Optional[bool] = None,
        serializer: Optional[str] = 'json',
        serializer_kwargs: Optional[Dict[str, Any]] = None,
        expiration_backend: Optional[Literal['auto', 'file', 'redis']] = 'auto', # ``
        **kwargs,
    ):
        """
        Initializes the backend

        Args:
            base_key (str, File): 
                The base key for the backend.
                Either a string or a file-io File object.

                Examples:
                - base_key = 's3://mybucket/mykey'
                - base_key = File('s3://mybucket/mykey')

            name (str, optional): The name of the backend. Defaults to None.
            expiration (int, optional): The expiration time in seconds. Defaults to None.
        """
        if isinstance(base_key, str): base_key = File(base_key)
        self.base_key: 'File' = base_key
        if name is not None: self.name = name
        if expiration is not None: 
            self.expiration = expiration
        self.async_enabled = async_enabled
        self.expiration_backend = expiration_backend

        if file_ext is not None: 
            file_ext = file_ext.lstrip('.')
            self.file_ext = file_ext
        if file_pre is not None: self.file_pre = file_pre
        if auto_delete_invalid is not None: self.auto_delete_invalid = auto_delete_invalid
        self._setup_exp_backend()

        # self.exp_file: 'File' = self.base_key.joinpath(f'.{self.name}.metadata.expires')
        
        from lzl.io.ser import get_serializer
        self.serializer = get_serializer(serializer = serializer, **(serializer_kwargs or {}))
        self._exp_ser = get_serializer(serializer = 'json')
        self._kwargs = kwargs
        self._kwargs['serializer'] = serializer
        self._kwargs['serializer_kwargs'] = serializer_kwargs
        self._kwargs['async_enabled'] = async_enabled
        self._kwargs['expiration'] = expiration
        self._kwargs['name'] = name
        self._kwargs['file_ext'] = file_ext
        self._kwargs['file_pre'] = file_pre
        self._kwargs['auto_delete_invalid'] = auto_delete_invalid
        self._kwargs['expiration_backend'] = self.expiration_backend
    
    def _setup_exp_backend(self):
        """
        Sets up the expiration backend
        """
        # Try to determine the expiration backend
        # Also handle automigration
        if self.expiration_backend == 'auto': 
            if RedisExpirationBackend.is_available(): self.expiration_backend = 'redis'
            else: self.expiration_backend = 'file'
            _display_backend(self.expiration_backend)
        if self.expiration_backend == 'file':
            self.exp_backend = FileExpirationBackend(backend = self)
        elif self.expiration_backend == 'redis':
            self.exp_backend = RedisExpirationBackend(backend = self)
        else:
            raise ValueError(f'Invalid Expiration Backend: {self.expiration_backend}')
        # logger.info(f'Using Expiration Backend: {self.exp_backend.name}')
        

    def get_writer(self, f: 'File', is_async: Optional[bool] = None) -> Callable[[Union[str, bytes]], Any]:
        """
        Returns a writer for the file
        """
        if is_async:
            return f.awrite_bytes if self.serializer.is_binary else f.awrite_text
        return f.write_bytes if self.serializer.is_binary else f.write_text

    def read_file(self, f: 'File', is_async: Optional[bool] = None):
        """
        Returns a reader for the file
        """
        if is_async:
            return f.aread_bytes() if self.serializer.is_binary else f.aread_text()
        return f.read_bytes() if self.serializer.is_binary else f.read_text()

    def write_file(self, f: 'File', data: Any, is_async: Optional[bool] = None):
        """
        Returns a writer for the file
        """
        if is_async:
            return f.awrite_bytes(data) if self.serializer.is_binary else f.awrite_text(data)
        return f.write_bytes(data) if self.serializer.is_binary else f.write_text(data)

    def _precheck(self, **kwargs):
        """
        Run a precheck operation
        """
        pass


    """
    Implemented Methods
    """

    def get_key(self, key: str) -> 'FileLike':
        """
        Gets a Key
        """
        if self.file_pre: key = self.file_pre + key
        if self.file_ext: key += f'.{self.file_ext}'
        return self.base_key.joinpath(key)
    
    def _fetch_one(
        self, 
        key: str, 
        default: Optional[Any] = None, 
        _raw: Optional[bool] = None, 
        _with_key: Optional[bool] = False,
        **kwargs
    ) -> Union[Optional[Any], Tuple[str, Any]]:
        """
        Fetches a Value from the Object Store
        """
        f_key = self.get_key(key)
        if not f_key.exists(): 
            return (key, default) if _with_key else default
        value = self.read_file(f_key)
        if value is None: 
            return (key, default) if _with_key else default
        if _raw: return (key, value) if _with_key else value
        try:
            decoded = self.decode_value(value, _raw = _raw, **kwargs)
            if decoded is None: 
                return (key, default) if _with_key else default
            return (key, decoded) if _with_key else decoded
        except Exception as e:
            logger.info(f'Error Decoding Value: |r|({type(value)}) {e}|e| {value}', colored = True, prefix = f_key.as_posix())
            if self.auto_delete_invalid: self.delete(key)
            return (key, default) if _with_key else default
    
    def _set_one(
        self, 
        key: str, 
        value: Any,
        _raw: Optional[bool] = None, 
        _with_key: Optional[bool] = False,
        **kwargs
    ) -> Union[Optional['File'], Tuple[str, 'File']]:
        """
        Fetches a Value from the Object Store
        """
        f_key = self.get_key(key)
        try:
            encoded = self.encode_value(value, _raw = _raw, **kwargs)
        except Exception as e:
            logger.info(f'Error Encoding Value: |r|({type(value)}) {e}|e| {value}', colored = True, prefix = f_key.as_posix())
            return (key, None) if _with_key else None
        try:
            self.write_file(f_key, encoded)
            return (key, f_key) if _with_key else f_key
        except Exception as e:
            logger.info(f'Error Writing Value: |r|({type(value)}) {e}|e| {value}', colored = True, prefix = f_key.as_posix())
            return (key, None) if _with_key else None


    def get(self, key: str, default: Optional[Any] = None, _raw: Optional[bool] = None, **kwargs) -> Optional[Any]:
        """
        Gets a Value from the Object Store
        """
        self.exp_backend._check(key)
        # self._run_expiration_check(key)
        return self._fetch_one(key, default, _raw = _raw, **kwargs)

    def get_values(self, keys: Iterable[str]) -> List[Any]:
        """
        Gets a Value from the Object Store
        """
        # self._run_expiration_check(*keys)
        self.exp_backend._check(*keys)
        results = []
        result_map = {key: None for key in keys}
        with concurrent.futures.ThreadPoolExecutor() as executor:
            futures = [executor.submit(self._fetch_one, key, _with_key = True) for key in keys]
            results.extend(
                future.result()
                for future in concurrent.futures.as_completed(futures)
            )
        # Map them to the original index
        for key, value in results:
            result_map[key] = value
        return list(result_map.values())
        

    def set(self, key: str, value: Any, ex: Optional[int] = None, _raw: Optional[bool] = None, **kwargs) -> Optional['File']:
        """
        Saves a Value to the Object Store
        """
        f_key = self._set_one(key, value, _raw = _raw, **kwargs)
        if f_key is None: return
        self.exp_backend._set(key, ex = ex)
        # self._set_expiration(key, ex = ex)
        return f_key

    def _set_batch(self, data: Dict[str, Any], ex: Optional[int] = None, _raw: Optional[bool] = None, **kwargs) -> Dict[str, Optional['File']]:
        """
        Saves a Value to the Object Store
        """
        results = []
        exp_keys = []
        result_map = {key: None for key in data}
        with concurrent.futures.ThreadPoolExecutor() as executor:
            futures = [executor.submit(self._set_one, key, value, _with_key = True, _raw = _raw, **kwargs) for key, value in data.items()]
            results.extend(
                future.result()
                for future in concurrent.futures.as_completed(futures)
            )
        # Map them to the original index
        for key, value in results:
            result_map[key] = value
            if value is not None: exp_keys.append(key)
        if exp_keys: self.exp_backend._set(*exp_keys, ex = ex)
        # if exp_keys: self._set_expiration(*exp_keys, ex = ex)
        return result_map

    def _set_batch_fallback(self, data: Dict[str, Any], ex: Optional[int] = None, _raw: Optional[bool] = None, **kwargs) -> Dict[str, Optional['File']]:
        """
        Saves a Value to the Object Store
        """
        exp_keys = []
        result_map = {key: None for key in data}
        for key, value in data.items():
            result_map[key] = self._set_one(key, value, _raw = _raw, **kwargs)
            if result_map[key] is not None: exp_keys.append(key)
        if exp_keys: self.exp_backend._set(*exp_keys, ex = ex)
        # if exp_keys: self._set_expiration(*exp_keys, ex = ex)
        return result_map

    def set_batch(self, data: Dict[str, Any], ex: Optional[int] = None, _raw: Optional[bool] = None, **kwargs) -> Dict[str, Optional['File']]:
        """
        Saves a Value to the Object Store
        """
        try:
            return self._set_batch(data, ex = ex, _raw = _raw, **kwargs)
        except RuntimeError as e:
            # logger.info(f'Unable to Set Batch: {e}', colored = True, prefix = self.base_key.as_posix())
            raise e
        except Exception as e:
            logger.info(f'[Fallback] Error Setting Batch: {e}', colored = True, prefix = self.base_key.as_posix())
            return self._set_batch_fallback(data, ex = ex, _raw = _raw, **kwargs)

    def _delete_one(self, key: str, **kwargs) -> None:
        """
        Deletes a Value from the Object Store
        """
        f_key = self.get_key(key)
        f_key.rm(missing_ok=True)

    def delete(self, key: str, **kwargs) -> None:
        """
        Deletes a Value from the Object Store
        """
        self.exp_backend._remove(key)
        self._delete_one(key, **kwargs)
        

    def _clear(self, *keys: str, **kwargs):
        """
        Clears the Keys from the Object Store
        """
        for _ in ThreadPool.iterate(
            self._delete_one,
            keys,
            return_ordered = False,
            **kwargs
        ):
            pass

        
    def clear(self, *keys: str, **kwargs):
        """
        Clears the Keys from the Object Store
        """
        # self._remove_expiration(*keys)
        self.exp_backend._remove(*keys)
        self._clear(*keys, **kwargs)

    async def _afetch_one(
        self, 
        key: str, 
        default: Optional[Any] = None, 
        _raw: Optional[bool] = None, 
        _with_key: Optional[bool] = False,
        **kwargs
    ) -> Union[Optional[Any], Tuple[str, Any]]:
        """
        Fetches a Value from the Object Store
        """
        f_key = self.get_key(key)
        if not await f_key.aexists(): 
            return (key, default) if _with_key else default
        value = await self.read_file(f_key, is_async = True)
        if value is None: 
            return (key, default) if _with_key else default
        if _raw: return (key, value) if _with_key else value
        try:
            decoded = await self.adecode_value(value, _raw = _raw, **kwargs)
            if decoded is None: 
                return (key, default) if _with_key else default
            return (key, decoded) if _with_key else decoded
        except Exception as e:
            logger.info(f'Error Decoding Value: |r|({type(value)}) {e}|e| {value}', colored = True, prefix = f_key.as_posix())
            if self.auto_delete_invalid: await self.adelete(key)
            return (key, default) if _with_key else default
    
    async def _aset_one(
        self, 
        key: str, 
        value: Any,
        _raw: Optional[bool] = None, 
        _with_key: Optional[bool] = False,
        **kwargs
    ) -> Union[Optional['File'], Tuple[str, 'File']]:
        """
        Fetches a Value from the Object Store
        """
        f_key = self.get_key(key)
        try:
            encoded = await self.aencode_value(value, _raw = _raw, **kwargs)
        except Exception as e:
            logger.info(f'Error Encoding Value: |r|({type(value)}) {e}|e| {value}', colored = True, prefix = f_key.as_posix())
            return (key, None) if _with_key else None
        try:
            await self.write_file(f_key, encoded, is_async = True)
            return (key, f_key) if _with_key else f_key
        except Exception as e:
            logger.info(f'Error Writing Value: |r|({type(value)}) {e}|e| {value}', colored = True, prefix = f_key.as_posix())
            return (key, None) if _with_key else None

    
    async def _aset_one_iter(
        self, 
        item: Tuple[str, Any],
        _raw: Optional[bool] = None, 
        _with_key: Optional[bool] = False,
        **kwargs
    ) -> Union[Optional['File'], Tuple[str, 'File']]:
        """
        Fetches a Value from the Object Store
        """
        key, value = item
        return await self._aset_one(key, value, _raw = _raw, _with_key = _with_key, **kwargs)

    async def _adelete_one(
        self,
        key: str,
        **kwargs,
    ):
        """
        Deletes a Value from the Object Store
        """
        f_key = self.get_key(key)
        await f_key.arm(missing_ok=True)

    async def aget(self, key: str, default: Optional[Any] = None, _raw: Optional[bool] = None, **kwargs) -> Optional[Any]:
        """
        Gets a Value from the Object Store
        """
        await self.exp_backend._acheck(key)
        return await self._afetch_one(key, default, _raw = _raw, **kwargs)

    async def aget_values(self, keys: Iterable[str], **kwargs) -> List[Any]:
        """
        Gets a Value from the DB
        """
        await self.exp_backend._acheck(*keys)
        results = []
        async for result in ThreadPool.aiterate(
            self._afetch_one,
            keys,
            return_ordered = True,
            **kwargs
        ):
            results.append(result)
        return results

    
    async def aset(self, key: str, value: Any, ex: Optional[int] = None, _raw: Optional[bool] = None, **kwargs) -> None:
        """
        Saves a Value to the DB
        """
        f_key = await self._aset_one(key, value, _raw = _raw, **kwargs)
        if f_key is None: return
        await self.exp_backend._aset(key, ex = ex)
        return f_key
        

    async def aset_batch(self, data: Dict[str, Any], ex: Optional[int] = None, **kwargs) -> None:
        """
        Saves a Value to the DB
        """
        results = []
        exp_keys = []
        items = list(data.items())
        async for (key, value) in ThreadPool.aiterate(
            self._aset_one_iter,
            items,
            return_ordered = True,
            _with_key = True,
            **kwargs
        ):
            results.append(value)
            if value is not None: exp_keys.append(key)
        if exp_keys: await self.exp_backend._aset(*exp_keys, ex = ex)
        # if exp_keys: await self._aset_expiration(*exp_keys, ex = ex)
        return results

    async def adelete(self, key: str, **kwargs) -> None:
        """
        Deletes a Value from the DB
        """
        await self.exp_backend._aremove(key)
        # await self._aremove_expiration(key)
        await self._adelete_one(key, **kwargs)

    async def _aclear(self, *keys: str, **kwargs):
        """
        Clears the Cache
        """
        await ThreadPool.amap(
            self._adelete_one,
            keys,
            return_ordered = False,
            **kwargs
        )

    async def aclear(self, *keys: str, **kwargs):
        """
        Clears the Cache
        """
        await self.exp_backend._aremove(*keys)
        # await self._aremove_expiration(*keys)
        await self._aclear(*keys, **kwargs)

    def _fetch_objstr_f_keys(
        self, 
        pattern: Optional[str] = '*', 
        **kwargs
    ) -> List['File']:
        """
        Returns the Keys within the current object storage
        """
        self.exp_backend._check(validate = True)
        return [f_key for f_key in self.base_key.glob(pattern = pattern) if f_key.is_file() and 'metadata.expires' not in f_key.name]

    async def _afetch_objstr_f_keys(
        self,
        pattern: Optional[str] = '*', 
        **kwargs
    ) -> List['File']:
        """
        Returns the Keys within the current object storage
        """
        await self.exp_backend._acheck(validate = True)
        f_keys: List['File'] = list(await self.base_key.aglob(pattern=pattern))
        return [f_key for f_key in f_keys if f_key.is_file() and 'metadata.expires' not in f_key.name]

    def _parse_f_keys_to_str(
        self,
        f_keys: List['File'],
        exclude_base_key: Optional[bool] = None,
        **kwargs
    ) -> List[str]:
        """
        Returns the Keys within the current object storage
        """
        f_key_strs: List[str] = [f_key.as_posix() for f_key in f_keys]
        if exclude_base_key: 
            base_key_str = self.base_key.as_posix()
            f_key_strs = [f_key.replace(base_key_str, '') for f_key in f_key_strs]
        return f_key_strs

    def get_keys(
        self, 
        pattern: str, 
        exclude_base_key: Optional[bool] = None, 
        as_file: Optional[bool] = None,
        **kwargs
    ) -> Union[List[str], List['File']]:
        """
        Returns the Keys within the current object storage
        """
        # self._run_expiration_check()
        self.exp_backend._check(validate = True)
        f_keys = self._fetch_objstr_f_keys(pattern = pattern)
        if as_file: return f_keys
        return self._parse_f_keys_to_str(f_keys, exclude_base_key = exclude_base_key, **kwargs)

    def get_all_keys(
        self, 
        exclude_base_key: Optional[bool] = False, 
        as_file: Optional[bool] = None,
        **kwargs
    ) -> Union[List[str], List['File']]:
        """
        Returns all the Keys
        """
        self.exp_backend._check(validate = True)
        f_keys = self._fetch_objstr_f_keys()
        if as_file: return f_keys
        return self._parse_f_keys_to_str(f_keys, exclude_base_key = exclude_base_key, **kwargs)
    

    def length(self, **kwargs) -> int:
        """
        Returns the Length of the Cache
        """
        # self._run_expiration_check()
        self.exp_backend._check(validate = True)
        f_keys = self._fetch_objstr_f_keys()
        return len(f_keys)
    
    async def alength(self, **kwargs) -> int:
        """
        Returns the Length of the Cache
        """
        # await self._arun_expiration_check()
        await self.exp_backend._acheck(validate = True)
        f_keys = await self._afetch_objstr_f_keys()
        return len(f_keys)

    def __len__(self):
        """
        Returns the Length of the Cache
        """
        return self.length()

    def contains(self, key, **kwargs):
        """
        Returns True if the Cache contains the Key
        """
        self.exp_backend._check(key)
        f_key = self.get_key(key)
        return f_key.exists()
    
    async def acontains(self, key, **kwargs):
        """
        Returns True if the Cache contains the Key
        """
        await self.exp_backend._acheck(key)
        f_key = self.get_key(key)
        return await f_key.aexists()
    
    def expire(self, key: str, ex: int, **kwargs) -> None:
        """
        Expires a Key
        """
        self.exp_backend._set(key, ex = ex, validate = True)
    
    async def aexpire(self, key: str, ex: int, **kwargs) -> None:
        """
        Expires a Key
        """
        await self.exp_backend._aset(key, ex = ex, validate = True)

    def purge(self, **kwargs) -> None:
        """
        Purges the cache
        """
        logger.warning('Purging the Cache is not recommended. Use `clear` instead.')
        f_keys = list(self.base_key.glob(pattern = '*', as_path = False))
        logger.info(f'Purging {len(f_keys)} Keys')
        self.base_key.rm(recursive = True, missing_ok = True)
        # self.clear(*f_keys, **kwargs)

    async def apurge(self, **kwargs) -> None:
        """
        Purges the cache
        """
        logger.warning('Purging the Cache is not recommended. Use `clear` instead.')
        f_keys = list(await self.base_key.aglob(pattern = '*', as_path = False))
        logger.info(f'Purging {len(f_keys)} Keys')
        await self.base_key.arm(recursive = True, missing_ok = True)
        # await self.aclear(*f_keys, **kwargs)


    """
    Add methods that reflect the `PersistentDict` API
    so that it can be used as a standalone backend
    """

    def get_child(self, key: str, **kwargs) -> 'ObjStorageStatefulBackend':
        """
        Gets a Child Persistent Dictionary
        """
        base_key = self.base_key.joinpath(key)
        if 'async_enabled' not in kwargs:
            kwargs['async_enabled'] = self.async_enabled
        base_kwargs = self._kwargs.copy()
        base_kwargs.update(kwargs)
        return self.__class__(
            base_key = base_key, 
            **base_kwargs
        )


    def __repr__(self):
        """
        Returns the Representation of the Cache
        """
        return f"<{self.__class__.__name__} num_keys={len(self)}, base_key={self.base_key}, serializer={self.serializer.name}>"


from __future__ import annotations

import typing as t
from ._pydantic import pydantic_data_converter
from ._pydantic_json import pydantic_data_converter as pydantic_json_data_converter
from ._serialized import serialized_data_converter, get_serialized_data_converter

if t.TYPE_CHECKING:
    from temporalio.converter import DataConverter


def get_default_data_converter(
    converter: t.Optional[str] = None,
    jsonlib: t.Optional[str] = None,
    compression: t.Optional[str] = None,
    compression_level: t.Optional[int] = None,
    **kwargs
) -> 'DataConverter':
    """
    Returns the default data converter
    """
    if converter == "serialized":
        return get_serialized_data_converter(jsonlib = jsonlib, compression = compression, compression_level = compression_level, **kwargs)
    if converter == "pydantic":
        return pydantic_data_converter
    if converter == "pydantic_json":
        return pydantic_json_data_converter
    from temporalio.converter import DataConverter
    return DataConverter.default
    # raise ValueError(f"Invalid converter: {converter}")


from .base import OpenAIClient
from .external import ExternalOpenAIClient, ExternalProviderSettings
from .loadbalancer import ClientLoadBalancer
from .manager import OpenAIManager
from .functions import FunctionManager, OpenAIFunctions
from lzl.proxied import ProxyObject

OpenAI: OpenAIManager = ProxyObject(OpenAIManager)



from __future__ import annotations

import typing as t
from io import BufferedReader # type: ignore
from lzl.types import BaseModel, Field, Literal

class ExtractousResult(BaseModel):
    """
    Extractous Result
    """
    result: str = Field(None, description = "The extracted result")
    metadata: t.Dict[str, t.Union[t.List[str], t.Any]] = Field(default_factory=dict, description = "The extracted metadata")

    @property
    def text(self) -> str:
        """
        Returns the text for the result
        """
        return self.result

    @property
    def url(self) -> t.Optional[str]:
        """
        Returns the URL
        """
        return self.metadata.get('url')
    
    @property
    def content_type(self) -> t.Optional[str]:
        """
        Returns the content type / mime type
        """
        if self.metadata.get('Content-Type'):
            return self.metadata['Content-Type'][0]
        if self.metadata.get('dc:format'):
            return self.metadata['dc:format'][0].split(';', 1)[0].strip()
        return None
    
    @property
    def filename(self) -> t.Optional[str]:
        """
        Returns the filename
        """
        if self.metadata.get('filename'):
            return self.metadata['filename'][0]
        if self.metadata.get('resourceName'):
            return self.metadata['resourceName'][0]
        if self.metadata.get('Content-Disposition'):
            return self.metadata['Content-Disposition'][0].split(';', 1)[1].split('=', 1)[1].strip()
        return None

    @property
    def num_pages(self) -> t.Optional[int]:
        """
        Returns the number of pages
        """
        if self.metadata.get('num_pages'):
            return int(self.metadata['num_pages'][0])
        if self.metadata.get('xmpTPg:NPages'):
            return int(self.metadata['xmpTPg:NPages'][0])
        return None
    
    @property
    def content_length(self) -> t.Optional[int]:
        """
        Returns the content length
        """
        if self.metadata.get('Content-Length'):
            return int(self.metadata['Content-Length'][0])
        return None
    
    @property
    def created_at(self) -> t.Optional[str]:
        """
        Returns the created at in string format
        """
        if self.metadata.get('created_at'):
            return self.metadata['created_at'][0]
        if self.metadata.get('dcterms:created'):
            return self.metadata['dcterms:created'][0]
        if self.metadata.get('pdf:docinfo:created'):
            return self.metadata['pdf:docinfo:created'][0]
        return None

    @property
    def modified_at(self) -> t.Optional[str]:
        """
        Returns the modified at in string format
        """
        if self.metadata.get('modified_at'):
            return self.metadata['modified_at'][0]
        if self.metadata.get('dcterms:modified'):
            return self.metadata['dcterms:modified'][0]
        if self.metadata.get('pdf:docinfo:modified'):
            return self.metadata['pdf:docinfo:modified'][0]
        return None


    def __str__(self) -> str:
        """
        Returns the string representation of the result
        """
        return self.result or ''

ExtractorMode = Literal['raw', 'dict', 'string', 'object']
ExtractorResult = t.Tuple[BufferedReader | t.Dict[str, t.Union[t.List[str], t.Any]]] |  str | t.Dict[str, t.Union[t.Dict[str, t.Any], t.List[str], t.Any]] | ExtractousResult
PDFOCRStrategy = Literal['NO_OCR', 'OCR_ONLY', 'OCR_AND_TEXT_EXTRACTION', 'AUTO']



from __future__ import annotations

from typing import Dict, Any, Optional
from lzl.proxied import ProxyObject
from .openai import OpenAISettings, AzureOpenAISettings
from .external import ExternalProviderSettings
from .proxy import (
    OpenAIProxySettings, 
    BaseProxySettings, 
    HeliconeProxySettings,
)

_openai_settings: Optional['OpenAISettings'] = None

def get_settings(**kwargs) -> OpenAISettings:
    """
    Returns the OpenAI Settings
    """
    global _openai_settings
    if _openai_settings is None:
        _openai_settings = OpenAISettings()
    if kwargs: _openai_settings.configure(**kwargs)
    return _openai_settings

def get_default_headers() -> Dict[str, Any]:
    """
    Returns the Default Headers
    """
    return get_settings().get_headers()

def get_max_retries() -> int:
    """
    Returns the Max Retries
    """
    return get_settings().max_retries


settings: OpenAISettings = ProxyObject(
    obj_getter = get_settings
)



from __future__ import annotations

"""Primary entry-point for LazyOps file abstractions.

The implementations stay untouched; only documentation and typing surface areas
are clarified to make downstream usage explicit for auto-generated docs.
"""

import os
import typing as t
from lzl.types import PYDANTIC_VERSION
if PYDANTIC_VERSION == 2:
    from pydantic import TypeAdapter

if t.TYPE_CHECKING:
    from .spec.main import FileLike, FileLikeT, PathLike, CloudFileSystemPath
    from .configs.main import FileIOConfig
    from .types.misc import ObjectSize


if t.TYPE_CHECKING:
    
    class File(t.Generic[FileLikeT], CloudFileSystemPath):
        """Lightweight wrapper over provider-specific path implementations."""
        
        settings: 'FileIOConfig'

        @classmethod
        def get_dir(cls, path: PathLike) -> FileLike:
            """
            Returns the directory of the file
            """
            ...



else:
    from .path import Path
    from .types.base import FilePath
    from .spec.paths.aws import FileS3Path
    from .spec.paths.minio import FileMinioPath
    from .utils.registry import fileio_settings

    # TypeAdapter(FileMinioPath)
    # TypeAdapter(FileS3Path)

    PathLike = t.TypeVar(
        'PathLike', 
        bound = t.Union[
            str, 
            os.PathLike[str],
            Path,
            FilePath,
            FileS3Path,
            FileMinioPath,
        ]
    )
    
    FileLike = t.Union[
        Path,
        FilePath,
        FileS3Path,
        FileMinioPath,
    ]
    
    FileLikeT = t.TypeVar(
        'FileLikeT',
        Path,
        FilePath,
        FileS3Path,
        FileMinioPath,
    )

    class File(t.Generic[FileLikeT]):
        """Factory that instantiates concrete path objects for various backends."""

        settings: 'FileIOConfig' = fileio_settings

        @classmethod
        def _get_filelike(cls, *args: t.Any, **kwargs: t.Any) -> 'FileLike':
            """Resolve the correct :class:`FileLike` instance for given input.

            Args:
                *args: Positional arguments forwarded to
                    :func:`lzl.io.file.spec.main.get_filelike`.
                **kwargs: Keyword arguments forwarded to the same helper.
            """
            from .spec.main import get_filelike
            return get_filelike(*args, **kwargs)
        
        @classmethod
        def get_object_size(cls, obj: t.Any) -> 'ObjectSize':
            """Return a convenience wrapper reporting object size in bytes."""
            from .types.misc import ObjectSize
            return ObjectSize(obj)

        def __new__(
            cls, 
            *args, 
            **kwargs
        ) -> 'FileLike':
            """Instantiate a file path object for the configured backend."""
            return cls._get_filelike(*args, **kwargs)


        @classmethod
        def get_dir(cls, path: 'PathLike') -> 'FileLike':
            """Return the parent directory for the provided path-like value."""
            new = cls._get_filelike(path)
            return new if new.is_dir() else new.parent
        
            
        @classmethod
        def register_loader(cls, ext: str, loader: t.Union[t.Callable[['FileLike'], None], t.Awaitable['FileLike', None]], overwrite: t.Optional[bool] = None) -> None:
            """Register a loader callback for the given file extension.

            Args:
                ext: Extension (``.json``, ``.csv``…) to register the loader
                    against.  A leading dot is optional.
                loader: Callable that receives the resolved :class:`FileLike`
                    instance and should return either a processed value or
                    coroutine.
                overwrite: When ``True`` the loader replaces any existing
                    registration for ``ext``.
            """
            from lzl.io.file.registry import register_loader
            register_loader(ext, loader, overwrite)

        # Pydantic methods
        if PYDANTIC_VERSION == 2:
            from pydantic_core import core_schema, SchemaSerializer
            from pydantic.annotated_handlers import GetCoreSchemaHandler, GetJsonSchemaHandler
            from pydantic.json_schema import JsonSchemaValue

            @classmethod
            def __get_pydantic_json_schema__(
                cls, core_schema: core_schema.CoreSchema, handler: GetJsonSchemaHandler
            ) -> JsonSchemaValue:
                
                field_schema = handler(core_schema)
                field_schema.update(format = 'path', type = 'string')
                return field_schema

            @classmethod
            def __get_pydantic_core_schema__(
                cls, 
                source: type[t.Any], 
                handler: GetCoreSchemaHandler
            ) -> core_schema.CoreSchema:
                """Return the Pydantic v2 CoreSchema for :class:`File` fields."""
                from pydantic_core import core_schema, SchemaSerializer
                schema = core_schema.with_info_plain_validator_function(
                    cls._validate,
                    serialization = core_schema.to_string_ser_schema(),
                )
                cls.__pydantic_serializer__ = SchemaSerializer(schema)
                return schema
            

            @classmethod
            def _validate(cls, __input_value: t.Any, _: core_schema.ValidationInfo) -> FileLike:
                """Pydantic validator that coerces inputs into :class:`File` objects."""
                return cls._get_filelike(__input_value) if __input_value is not None else None
            

            def __hash__(self: FileLike) -> int:
                return hash(self.as_posix())


        else:
            @classmethod
            def __get_validators__(cls):
                yield cls.validate

            @classmethod
            def validate(cls, v: t.Union[FileLike, t.Any]) -> FileLike:
                return cls._get_filelike(v) if v is not None else None
            
            @classmethod
            def __modify_schema__(cls, field_schema: t.Dict[str, t.Any]) -> None:
                field_schema.update(
                    type='string',
                    format='binary',
                )


if PYDANTIC_VERSION == 2:
    TypeAdapter(File)
    



from __future__ import annotations

"""Facades for LazyOps proxy helpers used across the codebase."""

from .base import ProxyObject, ProxyObjT
from .extra import LockedSingleton, Singleton
from .wraps import proxied

__all__ = [
    "ProxyObject",
    "ProxyObjT",
    "Singleton",
    "LockedSingleton",
    "proxied",
]



from __future__ import annotations

"""
Base Hatchet Client with Modifications
"""
import os
import abc
import contextlib
from pathlib import Path
from lzl.logging import logger
from lzo.types import AppEnv
from lzl.proxied import proxied, ProxyObject
from .utils import get_hatchet_settings, set_hatchet_settings
from .version import CURRENT_VERSION as VERSION
from typing import Any, Dict, List, Optional, Union, Iterator, Callable, TYPE_CHECKING, ParamSpec, Type, TypeVar, Tuple, overload

if TYPE_CHECKING:
    
    from lzl.logging import Logger
    from lzl.api.hatchet.config import HatchetSettings
    from .worker import Worker
    from .context import Context
    from .session import HatchetSession
    from .workflow import WorkflowT
    from .workflow.ctx import WorkflowContext

    from hatchet_sdk.loader import ClientConfig
    from hatchet_sdk.workflow import WorkflowMeta
    from hatchet_sdk.rate_limit import RateLimit
    from hatchet_sdk.clients.rest.models import WorkflowRun
    from hatchet_sdk.client import ClientImpl
    from hatchet_sdk.workflows_pb2 import StickyStrategy


ResponseT = TypeVar('ResponseT')

RT = TypeVar('RT')
P = ParamSpec("P")

# @proxied
class HatchetClient(abc.ABC):
    """
    The Hatchet Client that proxies and manages multiple Hatchet Sessions
    """
    name: str = 'hatchet'
    kind: str = 'client'
    default_instance: str = 'default'

    @overload
    def __init__(
        self,
        settings: Optional['HatchetSettings'] = None,
        grpc_endpoint: Optional[str] = None,
        api_endpoint: Optional[str] = None,
        session_class: Optional[Union[Type['HatchetSession'], str]] = None,
        context_class: Optional[Union[Type['Context'], str]] = None,
        worker_class: Optional[Union[Type['Worker'], str]] = None,
        workflow_meta_class: Optional[Union[Type['WorkflowMeta'], str]] = None,
        workflow_context_class: Optional[Union[Type['WorkflowContext'], str]] = None,
        **kwargs,
    ):
        """
        Initializes the Hatchet Client
        """
        ...

    def __init__(
        self,
        settings: Optional['HatchetSettings'] = None,
        **kwargs,
    ):
        """
        Initializes the Hatchet Client
        """
        if settings is None: settings = get_hatchet_settings()
        else: set_hatchet_settings(settings)
        self.hsettings = settings
        from .state import GlobalHatchetContext, GlobalHatchetContextObject
        self.c: GlobalHatchetContextObject = GlobalHatchetContext
        self.configure_classes(**kwargs)
        self.sessions = self.c.sessions
        self.hsettings.configure(**kwargs)

    def configure_classes(
        self,
        worker_class: Optional[Union[Type['Worker'], str]] = None,
        context_class: Optional[Union[Type['Context'], str]] = None,
        session_class: Optional[Union[Type['HatchetSession'], str]] = None,
        workflow_meta_class: Optional[Union[Type['WorkflowMeta'], str]] = None,
        workflow_context_class: Optional[Union[Type['WorkflowContext'], str]] = None,
        **kwargs,
    ):
        """
        Configures the global classes that are used to initialize the Hatchet Sessions
        """
        return self.c.configure_classes(
            worker_class = worker_class,
            context_class = context_class,
            session_class = session_class,
            workflow_meta_class = workflow_meta_class,
            workflow_context_class = workflow_context_class,
            **kwargs,
        )
        
    
    def register_workflow(
        self,
        workflow_name: Optional[str] = None,
        workflow_obj: Optional[Union[str, Type['WorkflowT']]] = None,
        workflow_mapping: Optional[Dict[str, Union[str, Type['WorkflowT']]]] = None,
        instance: Optional[str] = None,
        overwrite: Optional[bool] = None,
        **kwargs,
    ):
        """
        Registers workflows to the global context
        """
        return self.c.register_workflow(
            workflow_name = workflow_name,
            workflow_obj = workflow_obj,
            workflow_mapping = workflow_mapping,
            instance = instance,
            overwrite = overwrite,
            **kwargs,
        )
    
    def configure(
        self,
        overwrite: Optional[bool] = None,
        **kwargs,
    ):
        """
        Configures the global settings
        """
        self.hsettings.configure(**kwargs)
        if overwrite is True: self.init_session(overwrite = overwrite, **kwargs)

    @property
    def worker_class(self) -> Type['Worker']:
        """
        Returns the worker class
        """
        return self.c.worker_class
    
    @property
    def context_class(self) -> Type['Context']:
        """
        Returns the context class
        """
        return self.c.context_class
    
    @property
    def session_class(self) -> Type['HatchetSession']:
        """
        Returns the session class
        """
        return self.c.session_class
    
    @property
    def workflow_meta_class(self) -> Type['WorkflowMeta']:
        """
        Returns the workflow meta class
        """
        return self.c.workflow_meta_class

    """
    Session Management
    """

    def create_session(
        self,
        instance: Optional[str] = None,
        config: Union['ClientConfig', Dict[str, Any]] = None,
        settings: Optional['HatchetSettings'] = None,
        **kwargs,
    ) -> 'HatchetSession':
        """
        Returns a new Hatchet Session

        - Does not perform validation with the current contexts
        """
        settings = settings or self.hsettings
        if instance is None: instance = self.default_instance
        return self.session_class(
            instance = instance,
            config = config,
            settings = settings,
            _ctx = self.c,
            **kwargs,
        )

    def session(
        self,
        instance: Optional[str] = None,
        config: Union['ClientConfig', Dict[str, Any]] = None,
        settings: Optional['HatchetSettings'] = None,
        set_as_ctx: Optional[bool] = None,
        overwrite: Optional[bool] = None,
        **kwargs,
    ) -> 'HatchetSession':
        """
        Initializes a Hatchet Session that is managed by the Hatchet Session Manager
        """
        if instance is None: instance = self.default_instance
        if instance in self.sessions and overwrite is not True:
            if set_as_ctx is True: self.c.set_ctx(instance = instance)
            return self.sessions[instance]
    
        session = self.create_session(instance = instance, config = config, settings = settings, **kwargs)
        if set_as_ctx is None: set_as_ctx = not len(self.sessions)
        self.sessions[instance] = session
        if set_as_ctx is True: self.c.set_ctx(instance = instance)
        return session

    def init_session(
        self,
        instance: Optional[str] = None,
        config: Union['ClientConfig', Dict[str, Any]] = None,
        settings: Optional['HatchetSettings'] = None,
        set_as_ctx: Optional[bool] = None,
        overwrite: Optional[bool] = None,
        **kwargs,
    ) -> 'HatchetSession':
        """
        Initializes a Hatchet Session that is managed by the Hatchet Session Manager

        - Raises an error if the session already exists
        """
        if instance is None: instance = self.default_instance
        if instance in self.sessions and overwrite is not True:
            raise ValueError(f'Session {instance} already exists. Set `overwrite` to True to overwrite')
        return self.session(instance = instance, config = config, settings = settings, set_as_ctx = set_as_ctx, **kwargs)
    
    def get_session(
        self,
        instance: Optional[str] = None,
        **kwargs,
    ) -> 'HatchetSession':
        """
        Returns the Hatchet Session with the given name
        """
        if instance is None: instance = self.c.current
        if instance is None: instance = self.default_instance
        if instance not in self.sessions:
            return self.session(instance = instance, **kwargs)
        return self.sessions[instance]


    def set_session(
        self,
        instance: str,
        **kwargs,
    ):
        """
        Set the current Hatchet Session
        """
        if instance not in self.sessions: raise ValueError(f'Invalid session instance: {instance}')
        self.c.set_ctx(instance = instance)    


    def add_session(
        self,
        session: 'HatchetSession',
        overwrite: Optional[bool] = None,
        set_as_ctx: Optional[bool] = None,
        **kwargs,
    ):
        """
        Adds a Hatchet Session to the Hatchet Session Manager
        """
        if not isinstance(session, self.session_class): raise ValueError(f'Invalid session type: {type(session)}')
        if session.instance in self.sessions and overwrite is not True:
            raise ValueError(f'The session with instance {session.instance} already exists. Use overwrite = True to overwrite the session')
        if set_as_ctx is None: set_as_ctx = not len(self.sessions)
        self.sessions[session.instance] = session
        if set_as_ctx is True: self.c.set_ctx(instance = session.instance)

    """
    Properties
    """

    @property
    def ctx(self) -> 'HatchetSession':
        """
        Returns the current session
        """
        if not self.c.ctx:
            self.session()
        return self.c.ctx
    
    @property
    def current(self) -> Optional[str]:
        """
        Returns the current session
        """
        return self.c.current or self.default_instance
    
    @property
    def client(self) -> 'ClientImpl':
        """
        Returns the client
        """
        return self.ctx.client
    
    @contextlib.contextmanager
    def with_session(self, instance: str) -> Iterator['HatchetSession']:
        """
        Returns the session with the given name as the current session
        """
        if instance not in self.sessions: 
            raise ValueError(f'Invalid session instance: `{instance}`. Initialize the session first using `session` or `init_session`')
        yield self.sessions[instance]

    """
    Session Wrap Methods
    """
    def _session_function(self, *args, _function: Optional[str] = None, session_ctx: Optional[str] = None, **kwargs) -> Union[ResponseT, Awaitable[ResponseT]]:
        """
        Wraps the session function
        """
        if session_ctx is None: session_ctx = self.current
        with self.with_session(session_ctx) as session:
            return getattr(session, _function)(*args, **kwargs)
    
    """
    Misc Properties
    """

    @property
    def logger(self) -> 'Logger':
        """
        Gets the logger
        """
        return self.hsettings.logger
    
    @property
    def autologger(self) -> 'Logger':
        """
        Gets the autologger
        """
        return self.hsettings.autologger
    
    """
    Passthrough Methods
    """


    def _get_workflow(self, op_name: str, version: Optional[str] = None) -> Type['WorkflowT']:
        """
        Gets the workflow
        """
        return self.ctx._get_workflow(op_name = op_name, version = version)


    def get_worker(self, name: str, max_runs: int | None = None) -> Worker:
        """
        Creates a new worker
        """
        return self.ctx.get_worker(name = name, max_runs = max_runs)
    

    def bind_step(
        self,
        obj: 'WorkflowT',
        func: Callable[P, RT],
        name: str = "",
        timeout: str = "60m",
        parents: List[str] = [],
        retries: int = 0,
        rate_limits: List['RateLimit'] | None = None,
    ):  # sourcery skip: default-mutable-arg
        """
        Binds a step to a workflow
        """
        return self.ctx.bind_step(
            obj = obj,
            func = func,
            name = name,
            timeout = timeout,
            parents = parents,
            retries = retries,
            rate_limits = rate_limits,
        )

    def bind_failure_step(
        self,
        obj: 'WorkflowT',
        func: Callable[P, RT],
        name: str = "",
        timeout: str = "",
        retries: int = 0,
        rate_limits: List['RateLimit'] | None = None,
    ):
        """
        Binds a step to a workflow
        """
        return self.ctx.bind_failure_step(
            obj = obj,
            func = func,
            name = name,
            timeout = timeout,
            retries = retries,
            rate_limits = rate_limits,
        )

    def build_workflow(
        self,
        obj: Type['WorkflowT'],
        name: str = "",
        on_events: List[str] = [],
        on_crons: list = [],
        version: str = "",
        timeout: str = "60m",
        schedule_timeout: str = "15m",
        sticky: 'StickyStrategy' = None,
    ) -> Type['WorkflowT']:  # sourcery skip: default-mutable-arg
        """
        Binds the workflow component to the workflow
        """
        return self.ctx.build_workflow(
            obj = obj,
            name = name,
            on_events = on_events,
            on_crons = on_crons,
            version = version,
            timeout = timeout,
            schedule_timeout = schedule_timeout,
            sticky = sticky,
        )


    def bind_workflow(
        self,
        obj: 'WorkflowT',
        name: str = "",
        on_events: List[str] = [],
        on_crons: list = [],
        version: str = "",
        timeout: str = "60m",
        schedule_timeout: str = "15m",
        sticky: 'StickyStrategy' = None,
    ) -> Type['WorkflowT']:  # sourcery skip: default-mutable-arg
        """
        Binds the workflow component to the workflow
        """
        return self.ctx.bind_workflow(
            obj = obj,
            name = name,
            on_events = on_events,
            on_crons = on_crons,
            version = version,
            timeout = timeout,
            schedule_timeout = schedule_timeout,
            sticky = sticky,
        )


    def chain_workflow(
        self,
        ref: 'WorkflowT',
        with_meta: Optional[bool] = True,
        with_links: Optional[bool] = True,
        **kwargs
    ) -> 'WorkflowT':
        """
        Chains the workflow
        """
        return self.ctx.chain_workflow(ref = ref, with_meta = with_meta, with_links = with_links, **kwargs)
    
    def gather_workflow(
        self,
        ref: 'WorkflowT',
        with_meta: Optional[bool] = True,
        with_links: Optional[bool] = True,
        **kwargs
    ) -> 'WorkflowT':
        """
        Gathers the workflow
        """
        return self.ctx.gather_workflow(ref = ref, with_meta = with_meta, with_links = with_links, **kwargs)

    def gather_workflows(
        self,
        workflow_names: Optional[List[str]] = None,
        include: Optional[List[str]] = None,
        exclude: Optional[List[str]] = None,
        **kwargs,
    ) -> Dict[str, 'WorkflowT']:
        """
        Gathers the workflows for the registered workflows
        """
        return self.ctx.gather_workflows(workflow_names = workflow_names, include = include, exclude = exclude, **kwargs)
    
    def gather_workers(
        self,
        workflow_names: Optional[List[str]] = None,
        include: Optional[List[str]] = None,
        exclude: Optional[List[str]] = None,
        **kwargs,
    ) -> Dict[str, 'Worker']:
        """
        Gathers the workers for the registered workflows
        """
        return self.ctx.gather_workers(workflow_names = workflow_names, include = include, exclude = exclude, **kwargs)
    

    """
    Compile Method
    """

    def compile_workflow(
        self,
        ref: Type['WorkflowT'],
        **kwargs
    ) -> 'WorkflowT': 
        """
        Compiles the workflow
        """
        return self.ctx.compile_workflow(ref = ref, **kwargs)
    
    def compile_workflows(
        self,
        include: Optional[List[str]] = None,
        exclude: Optional[List[str]] = None,
        extra_workflows: Optional[Dict[str, 'WorkflowT']] = None,
        include_crontasks: Optional[bool] = None,
        only_crontasks: Optional[bool] = None,
        **kwargs
    ) -> Dict[str, 'WorkflowT']:
        """
        Compiles the workflows
        """
        return self.ctx.compile_workflows(include = include, exclude = exclude, extra_workflows = extra_workflows, include_crontasks = include_crontasks, only_crontasks = only_crontasks, **kwargs)


    def delete_all_workflows(
        self,
        include: Optional[List[str]] = None,
        exclude: Optional[List[str]] = None,
    ):
        """
        Deletes all workflows
        """
        return self.ctx.delete_all_workflows(include = include, exclude = exclude)

    async def retrieve_workflow_run(
        self,
        request_id: str,
        blocking: bool = None,
        raise_error: Optional[bool] = False,
        **kwargs,
    ) -> Optional['WorkflowRun']:
        """
        Retrieves the workflow run
        """
        return await self.ctx.retrieve_workflow_run(request_id = request_id, blocking = blocking, raise_error = raise_error, **kwargs)
    
    async def determine_if_workflow_is_active(
        self,
        request_id: str,
        raise_error: Optional[bool] = False,
        **kwargs,
    ) -> bool:
        """
        Determines if the workflow is active
        """
        return await self.ctx.determine_if_workflow_is_active(request_id = request_id, raise_error = raise_error, **kwargs)

    async def abort_workflow_run(
        self,
        request_id: str,
        raise_error: Optional[bool] = False,
        **kwargs,
    ) -> Optional['WorkflowRun']:
        """
        Aborts the workflow run
        """
        return await self.ctx.abort_workflow_run(request_id = request_id, raise_error = raise_error, **kwargs)


    async def poll_workflow_run(
        self,
        request_id: str,
        **kwargs,
    ) -> Tuple[bool, Union[Dict[str, Any], Any], 'WorkflowRun']:
        """
        Polls the workflow run
        """
        return await self.ctx.poll_workflow_run(request_id = request_id, **kwargs)

    @classmethod
    def configure_from_env_file(
        cls,
        env_file: Optional[Union[str, Path]] = None,
        app_env: Optional[Union[str, AppEnv]] = None,
        app_name: Optional[str] = None,
        endpoint_env: Optional[str] = None,
        **kwargs,
    ):
        """
        Configures the Hatchet Client from an Environment File
        """
        if env_file is None: return
        if isinstance(env_file, str): env_file = Path(env_file)
        if not env_file.exists(): return
        import yaml
        config_data: Dict[str, Dict[str, Dict[str, str]]] = yaml.safe_load(env_file.read_text())
        settings = get_hatchet_settings()
        # if app_env is None: app_env = settings.app_env.name
        # if app_env and isinstance(app_env, AppEnv): app_env = app_env.name
        # if app_env and not isinstance(app_env, str): app_env = app_env.name
        # if app_env: app_env = str(app_env).lower()
        logger.info(f'Configuring Hatchet Client from Environment File: {env_file}', prefix = f'Hatchet - {app_env}')
        endpoint_config = config_data.get('endpoints')
        if endpoint_config:
            endpoint_env = endpoint_env or app_env
            if endpoint_env and endpoint_env in endpoint_config:
                settings.endpoints = endpoint_config[endpoint_env]
            else: settings.endpoints = endpoint_config
        if app_name:
            apikeys = config_data.get('apikeys')
            if apikeys and app_name in apikeys:
                app_env = app_env or 'production'
                if app_env in apikeys[app_name]:
                    os.environ['HATCHET_CLIENT_TOKEN'] = apikeys[app_name][app_env]


    

Hatchet: HatchetClient = ProxyObject(
    # obj_cls = HatchetClient,
    obj_getter = 'lzl.api.hatchet.utils.get_hatchet_client',
)




from typing import (
    Any,
    Callable,
    Iterable,
    List,
    Literal,
    Optional,
    TypedDict,
    Union,
    Protocol,
    TYPE_CHECKING,
)
import dataclasses
import logging
from .utils import logger

if TYPE_CHECKING:
    from .result import BulkResult


class LogMethod(Protocol):
    def __call__(self, msg: str, *, exc_info: bool = False) -> Any:
        ...


QueryInfoRequestType = Literal[
    "execute-read",
    "execute-write",
    "executemany",
    "executeunified-readonly",
    "executeunified-write",
]


@dataclasses.dataclass
class QueryInfoLazy:
    """
    Potentially lazily initialized query information.
    """

    operations: Union[Iterable[str], Callable[[], Iterable[str]]]
    """The operations as a list of SQL queries. May be built lazily."""
    params: Union[Iterable[Iterable[Any]], Callable[[], Iterable[Iterable[Any]]]]
    """The parameters for each query in operations. May be built lazily."""

    request_type: Union[QueryInfoRequestType, Callable[[], QueryInfoRequestType]]
    """
    The type of request that was made, may be determined lazily. This is one of
    - "execute-read": A single SELECT or EXPLAIN query via execute() or explain()
    - "execute-write": A single non-SELECT query via execute()
    - "executemany": A single execute_many() request, which is always treated as write
    - "executeunified-readonly": A single execute_unified() request where all of the
        queries are SELECT or EXPLAIN
    - "executeunified-write": A single execute_unified() request where at least one
        of the queries is not SELECT or EXPLAIN
    """
    consistency: Literal["none", "weak", "strong"]
    """
    The consistency level that was used for the request. Only matters if the
    request is a read (execute-read or executeunified-readonly).
    """
    freshness: str
    """
    The minimum freshness to use, e.g., 5m. Only relevant for read queries at the
    none consistency level.
    """


@dataclasses.dataclass
class QueryInfo:
    """
    The information about a query as it is passed to the slow query log
    method, before augmenting information that can only be known when the
    request is made
    """

    operations: Iterable[str]
    """The operations as a list of SQL queries. May be built lazily."""
    params: Iterable[Iterable[Any]]
    """The parameters for each query in operations. May be built lazily."""

    request_type: Literal[
        "execute-read",
        "execute-write",
        "executemany",
        "executeunified-readonly",
        "executeunified-write",
    ]
    """
    The type of request that was made. This is one of
    - "execute-read": A single SELECT or EXPLAIN query via execute() or explain()
    - "execute-write": A single non-SELECT query via execute()
    - "executemany": A single execute_many() request, which is always treated as write
    - "executeunified-readonly": A single execute_unified() request where all of the
        queries are SELECT or EXPLAIN
    - "executeunified-write": A single execute_unified() request where at least one
        of the queries is not SELECT or EXPLAIN
    """
    consistency: Literal["none", "weak", "strong"]
    """
    The consistency level that was used for the request. Only matters if the
    request is a read (execute-read or executeunified-readonly).
    """
    freshness: str
    """
    The minimum freshness to use, e.g., 5m. Only relevant for read queries at the
    none consistency level.
    """


class SlowQueryLogMethod(Protocol):
    def __call__(
        self,
        info: QueryInfo,
        /,
        *,
        duration_seconds: float,
        host: str,
        response_size_bytes: int,
        started_at: float,
        ended_at: float,
    ) -> None:
        """Called to log a slow query. Provided the operations and parameters in the
        same format as executemany2, then all the relevant context about the query
        that exceeded the threshold.

        Args:
            info (QueryInfo): The information about the query.
            duration_seconds (float): How long it took between us initiating the
                request and us receiving the response, in seconds
            host (str): The host that the request was made to.
            response_size_bytes (int): The size of the response in bytes, as reported by the
                content-length header. 0 if the content-length header was not present in the
                response.
            started_at (float): The time that the request was initiated, in seconds since the epoch.
            ended_at (float): The time that the response was received, in seconds since the epoch.
        """
        ...


class LogMessageConfig(TypedDict):
    """Configures a single log message within rqlite."""

    enabled: bool
    """True if the message should be logged, False otherwise. If not
    present, assumed to be True
    """

    method: LogMethod
    """The function to call to log the message. If not present,
    then this will be set based on the level of the message.
    For example, a level of DEBUG implies that the method is
    effectively logging.debug.

    The method should support "exc_info=True" as a keyword argument.
    """

    level: int
    """The level of the message. If not present, assumed to be
    logging.DEBUG. The level does not have to be one of the default
    logging levels - the method will be a partial variant of logging.log
    with the level as the first argument.

    The level is ignored if the method is set.
    """

    max_length: Optional[int]
    """The approximate maximum length of the message. This may be
    implemented differently depending on which message is being
    configured. If not present assumed to be None, for no maximum
    length.
    """


class LevelOnlyMessageConfig(TypedDict):
    """Used to appease the type system when initializing a log message config
    using only a debug level.
    """

    enabled: bool
    """See LogMessageConfig"""
    level: int
    """See LogMessageConfig"""


class DisabledMessageConfig(TypedDict):
    """Used to appease the type system when initializing a log message config
    which is disabled, since the other arguments are not needed.
    """

    enabled: Literal[False]
    """See LogMessageConfig"""


class SlowQueryLogMessageConfig(TypedDict):
    """The configuration available for the slow query log message, which
    is a special case because it's expected that these messages will be
    sent to a more visible place and formatted in a special way.
    """

    enabled: bool
    """True if the message should be logged, False otherwise. If not
    present, assumed to be True
    """

    threshold_seconds: float
    """The threshold in seconds for a query to be considered slow. Can
    be set to for detailed timing information on every query.
    """

    method: SlowQueryLogMethod
    """The function to call to log the message. If not present,
    then this will be set based on the level of the message.
    For example, a level of DEBUG implies that the method is
    effectively logging.debug.

    The method should support "exc_info=True" as a keyword argument.
    """


ForgivingLogMessageConfigT = Union[
    LogMessageConfig,
    LevelOnlyMessageConfig,
    DisabledMessageConfig,
]


@dataclasses.dataclass(frozen=True)
class LogConfig:
    """Describes the configuration of rqlite's logging."""

    read_start: ForgivingLogMessageConfigT = dataclasses.field(
        default_factory=lambda: LevelOnlyMessageConfig(
            enabled=True, level=logging.DEBUG
        )
    )
    """Configures the message to log when cursor.execute
    is called with a SELECT query.
    """

    read_response: ForgivingLogMessageConfigT = dataclasses.field(
        default_factory=lambda: LevelOnlyMessageConfig(
            enabled=True, level=logging.DEBUG
        )
    )
    """Configures the message to log when we get the response
    from the server for a SELECT query.
    """

    read_stale: ForgivingLogMessageConfigT = dataclasses.field(
        default_factory=lambda: LevelOnlyMessageConfig(
            enabled=True, level=logging.DEBUG
        )
    )
    """Configures the message to log when we get a response from
    the server for a SELECT query, but the response indicates we
    must retry because the data is not sufficiently fresh. This
    occurs only on reads with the read consistency level "none".
    """

    write_start: ForgivingLogMessageConfigT = dataclasses.field(
        default_factory=lambda: LevelOnlyMessageConfig(
            enabled=True, level=logging.DEBUG
        )
    )
    """Configures the message to log when cursor.execute
    is called with a non-SELECT query.
    """

    write_response: ForgivingLogMessageConfigT = dataclasses.field(
        default_factory=lambda: LevelOnlyMessageConfig(
            enabled=True, level=logging.DEBUG
        )
    )
    """Configures the message to log when we get the response
    from the server for a non-SELECT query.
    """

    connect_timeout: ForgivingLogMessageConfigT = dataclasses.field(
        default_factory=lambda: LevelOnlyMessageConfig(
            enabled=True, level=logging.WARNING
        )
    )
    """Configures the message to log when a connection attempt
    to one of the host nodes fails.
    """

    hosts_exhausted: ForgivingLogMessageConfigT = dataclasses.field(
        default_factory=lambda: LevelOnlyMessageConfig(
            enabled=True, level=logging.CRITICAL
        )
    )
    """Configures the message to log when we are going to give
    up on a given query because we have exhausted all attempts on
    all nodes. This implies the cluster is unresponsive or we cannot
    reach the cluster.
    """

    non_ok_response: ForgivingLogMessageConfigT = dataclasses.field(
        default_factory=lambda: LevelOnlyMessageConfig(
            enabled=True, level=logging.WARNING
        )
    )
    """Configures the message to log when we get a response from
    the server that is not OK or is a redirect when one is not
    expected, such as when we have exceeded the maximum number of
    redirects.
    """

    slow_query: Union[
        SlowQueryLogMessageConfig, DisabledMessageConfig
    ] = dataclasses.field(default_factory=lambda: DisabledMessageConfig(enabled=False))
    """Configures the message to log when we get a response from
    the server, but that response takes longer than a certain
    threshold to arrive.
    """

    backup_start: ForgivingLogMessageConfigT = dataclasses.field(
        default_factory=lambda: LevelOnlyMessageConfig(enabled=True, level=logging.INFO)
    )
    """Configures the message to log when we start attempting a backup."""

    backup_end: ForgivingLogMessageConfigT = dataclasses.field(
        default_factory=lambda: LevelOnlyMessageConfig(enabled=True, level=logging.INFO)
    )
    """Configures the message to log when we finish attempting a backup."""


DISABLED_LOG_CONFIG = LogConfig(
    read_start=DisabledMessageConfig(enabled=False),
    read_response=DisabledMessageConfig(enabled=False),
    read_stale=DisabledMessageConfig(enabled=False),
    write_start=DisabledMessageConfig(enabled=False),
    write_response=DisabledMessageConfig(enabled=False),
    connect_timeout=DisabledMessageConfig(enabled=False),
    hosts_exhausted=DisabledMessageConfig(enabled=False),
    non_ok_response=DisabledMessageConfig(enabled=False),
    slow_query=DisabledMessageConfig(enabled=False),
    backup_start=DisabledMessageConfig(enabled=False),
    backup_end=DisabledMessageConfig(enabled=False),
)
"""The log configuration which disables all logging."""


def log(
    config: ForgivingLogMessageConfigT,
    msg_supplier: Callable[[Optional[int]], str],
    exc_info: bool = False,
) -> None:
    """Logs a message if the config is enabled.

    Args:
        config: The configuration of the message to log.
        msg_supplier: A function which returns the message to log.
            Passed the approximate length of the message if there is
            one.
        exc_info: True to pass the current exception to the logger,
            False not to.
    """
    if not config.get("enabled", True):
        return

    max_length = config.get("max_length", None)
    message = msg_supplier(max_length)

    method = config.get("method", None)
    if method is not None:
        if not exc_info:
            method(message)
        else:
            method(message, exc_info=True)
        return

    level = config.get("level", logging.DEBUG)
    logger.log(level, message, exc_info=exc_info)